"""Gather the structure and discover the content."""

import json
import os
import pathlib
from typing import Dict, List, Set, Tuple, Union

import yaml

from liitos import (
    DEFAULT_STRUCTURE_NAME,
    KEY_APPROVALS,
    KEY_BIND,
    KEY_CHANGES,
    KEY_LAYOUT,
    KEY_META,
    KEYS_REQUIRED,
    ENCODING,
    log,
)

PathLike = Union[str, pathlib.Path]

Approvals = Dict[str, Union[List[str], List[List[str]]]]
Assets = Dict[str, Dict[str, Dict[str, str]]]
Binder = List[str]
Changes = Dict[str, Union[List[str], List[List[str]]]]
Layout = Dict[str, str]
Meta = Dict[str, str]
Structure = Dict[str, List[Dict[str, str]]]
Targets = Set[str]
Facets = Dict[str, Targets]
Payload = Union[Approvals, Binder, Changes, Meta]
Verification = Tuple[bool, str]


def load_structure(path: PathLike = DEFAULT_STRUCTURE_NAME) -> Structure:
    """Load the structure information and content links from the YAML file per convention."""
    with open(path, 'rt', encoding=ENCODING) as handle:
        return yaml.safe_load(handle)  # type: ignore


def targets(structure: Structure) -> Targets:
    """Extract the targets from the given structure information item."""
    return set(target for target in structure)


def facets(structure: Structure) -> Facets:
    """Extract the facets per target from the given structure information item."""
    return {target: set(facet for facet_data in cnt for facet in facet_data) for target, cnt in structure.items()}


def assets(structure: Structure) -> Assets:
    """Map the assets to facets of targets."""
    return {t: {f: asset for fd in cnt for f, asset in fd.items()} for t, cnt in structure.items()}  # type: ignore


def verify_target(name: str, target_set: Targets) -> Verification:
    """Verify presence of target yielding predicate and message (in case of failure)."""
    return (True, '') if name in target_set else (False, f'target ({name}) not in {sorted(target_set)}')


def verify_facet(name: str, target: str, facet_map: Facets) -> Verification:
    """Verify presence of facet for target yielding predicate and message (in case of failure)."""
    if name in facet_map[target]:
        return True, ''
    return False, f'facet ({name}) of target ({target}) not in {sorted(facet_map[target])}'


def error_context(
    payload: Payload,
    label: str,
    facet: str,
    target: str,
    path: PathLike,
    err: Union[FileNotFoundError, KeyError, ValueError],
) -> Tuple[Payload, str]:
    """Provide harmonized context for the error situation as per parameters."""
    if isinstance(err, FileNotFoundError):
        return payload, f'{label} link not found at ({path}) or invalid for facet ({facet}) of target ({target})'
    if isinstance(err, KeyError):
        return [], f'{label} not found in assets for facet ({facet}) of target ({target})'
    if isinstance(err, ValueError):
        return [], f'{label} requires json or yaml format in assets for facet ({facet}) of target ({target})'
    raise NotImplementedError(f'error context not implemented for error ({err})')


def load_binder(facet: str, target: str, path: PathLike) -> Tuple[Binder, str]:
    """Yield the binder for facet of target from path and message (in case of failure)."""
    try:
        with open(path, 'rt', encoding=ENCODING) as handle:
            return [line.strip() for line in handle.readlines() if line.strip()], ''
    except FileNotFoundError as err:
        return error_context([], 'Binder', facet, target, path, err)  # type: ignore


def binder(facet: str, target: str, asset_struct: Assets) -> Tuple[Binder, str]:
    """Yield the binder for facet of target from link in assets and message (in case of failure)."""
    try:
        path = pathlib.Path(asset_struct[target][facet][KEY_BIND])
    except KeyError as err:
        return error_context([], 'Binder', facet, target, '', err)  # type: ignore
    return load_binder(facet, target, path)


def load_layout(facet: str, target: str, path: PathLike) -> Tuple[Meta, str]:
    """Yield the layout for facet of target from path and message (in case of failure)."""
    try:
        with open(path, 'rt', encoding=ENCODING) as handle:
            return yaml.safe_load(handle), ''
    except FileNotFoundError as err:
        return error_context({}, 'Metadata', facet, target, path, err)  # type: ignore


def layout(facet: str, target: str, asset_struct: Assets) -> Tuple[Meta, str]:
    """Yield the layout for facet of target from link in assets and message (in case of failure)."""
    try:
        path = pathlib.Path(asset_struct[target][facet][KEY_LAYOUT])
    except KeyError as err:
        return error_context({}, 'Layout', facet, target, '', err)  # type: ignore
    return load_layout(facet, target, path)


def load_meta(facet: str, target: str, path: PathLike) -> Tuple[Meta, str]:
    """Yield the metadata for facet of target from path and message (in case of failure)."""
    try:
        with open(path, 'rt', encoding=ENCODING) as handle:
            return yaml.safe_load(handle), ''
    except FileNotFoundError as err:
        return error_context({}, 'Metadata', facet, target, path, err)  # type: ignore


def meta(facet: str, target: str, asset_struct: Assets) -> Tuple[Meta, str]:
    """Yield the metadata for facet of target from link in assets and message (in case of failure)."""
    try:
        path = pathlib.Path(asset_struct[target][facet][KEY_META])
    except KeyError as err:
        return error_context({}, 'Metadata', facet, target, '', err)  # type: ignore
    return load_meta(facet, target, path)


def load_approvals(facet: str, target: str, path: PathLike) -> Tuple[Approvals, str]:
    """Yield the approvals for facet of target from path and message (in case of failure)."""
    if str(path).lower().endswith('json'):
        try:
            with open(path, 'rt', encoding=ENCODING) as handle:
                return json.load(handle), ''
        except FileNotFoundError as err:
            return error_context({}, 'Approvals', facet, target, path, err)  # type: ignore
    elif str(path).lower().endswith(('yaml', 'yml')):
        try:
            with open(path, 'rt', encoding=ENCODING) as handle:
                return yaml.safe_load(handle), ''
        except FileNotFoundError as err:
            return error_context({}, 'Approvals', facet, target, path, err)  # type: ignore

    return error_context({}, 'Approvals', facet, target, path, ValueError('json or yaml required'))  # type: ignore


def approvals(facet: str, target: str, asset_struct: Assets) -> Tuple[Approvals, str]:
    """Yield the approvals for facet of target from link in assets and message (in case of failure)."""
    try:
        path = pathlib.Path(asset_struct[target][facet][KEY_APPROVALS])
    except KeyError as err:
        return error_context({}, 'Approvals', facet, target, '', err)  # type: ignore
    return load_approvals(facet, target, path)


def load_changes(facet: str, target: str, path: PathLike) -> Tuple[Approvals, str]:
    """Yield the changes for facet of target from path and message (in case of failure)."""
    if str(path).lower().endswith('json'):
        try:
            with open(path, 'rt', encoding=ENCODING) as handle:
                return json.load(handle), ''
        except FileNotFoundError as err:
            return error_context({}, 'Changes', facet, target, path, err)  # type: ignore
    elif str(path).lower().endswith(('yaml', 'yml')):
        try:
            with open(path, 'rt', encoding=ENCODING) as handle:
                return yaml.safe_load(handle), ''
        except FileNotFoundError as err:
            return error_context({}, 'Changes', facet, target, path, err)  # type: ignore

    return error_context({}, 'Changes', facet, target, path, ValueError('json or yaml required'))  # type: ignore


def changes(facet: str, target: str, asset_struct: Assets) -> Tuple[Changes, str]:
    """Yield the changes for facet of target from link in assets and message (in case of failure)."""
    try:
        path = pathlib.Path(asset_struct[target][facet][KEY_CHANGES])
    except KeyError as err:
        return error_context({}, 'Changes', facet, target, '', err)  # type: ignore
    return load_changes(facet, target, path)


def verify_asset_keys(facet: str, target: str, asset_struct: Assets) -> Verification:
    """Verify presence of required keys for facet of target yielding predicate and message (in case of failure)."""
    if all(key in asset_struct[target][facet] for key in KEYS_REQUIRED):
        return True, ''
    return False, f'keys in {sorted(KEYS_REQUIRED)} for facet ({facet}) of target ({target}) are missing'


def verify_asset_links(facet: str, target: str, asset_struct: Assets) -> Verification:
    """Verify presence of asset links for facet of target yielding predicate and message (in case of failure)."""
    predicate, message = verify_asset_keys(facet, target, asset_struct)
    if not predicate:
        return predicate, message
    for key in KEYS_REQUIRED:
        link = pathlib.Path(asset_struct[target][facet][key])
        log.debug(f'  + verifying: {pathlib.Path.cwd() / link}')
        if not link.is_file() or not link.stat().st_size:
            return False, f'{key} asset link ({link}) for facet ({facet}) of target ({target}) is invalid'
    return True, ''


ASSET_KEY_ACTION = {
    KEY_APPROVALS: approvals,
    KEY_BIND: binder,
    KEY_CHANGES: changes,
    KEY_LAYOUT: layout,
    KEY_META: meta,
}


def verify_assets(facet: str, target: str, asset_struct: Assets) -> Verification:
    """Verify assets for facet of target yielding predicate and message (in case of failure)."""
    predicate, message = verify_asset_links(facet, target, asset_struct)
    if not predicate:
        return predicate, message
    for key, action in ASSET_KEY_ACTION.items():
        asset, message = action(facet, target, asset_struct)
        if not asset:
            return False, f'{key} asset for facet ({facet}) of target ({target}) is invalid'
    return True, ''


def prelude(
    doc_root: Union[str, pathlib.Path], structure_name: str, target_key: str, facet_key: str, command: str
) -> tuple[Structure, Assets]:
    """DRY."""
    doc_root = pathlib.Path(doc_root)
    idem = os.getcwd()
    os.chdir(doc_root)
    job_description = (
        f'facet ({facet_key}) of target ({target_key}) with structure map ({structure_name})'
        f' in document root ({doc_root}) coming from ({idem})'
    )
    log.info(f'executing prelude of command ({command}) for {job_description}')
    structure = load_structure(structure_name)
    asset_map = assets(structure)
    return structure, asset_map


def verify(
    doc_root: Union[str, pathlib.Path],
    structure_name: str,
    target_key: str,
    facet_key: str,
    options: dict[str, Union[bool, str]],
) -> int:
    """Drive the verification."""
    doc_root = pathlib.Path(doc_root)
    os.chdir(doc_root)
    facet = facet_key
    target = target_key
    structure_name = structure_name
    job_description = (
        f'facet ({facet}) for target ({target}) with structure map ({structure_name})' f' in document root ({doc_root})'
    )
    log.info(f'starting verification of {job_description}')
    structure = load_structure(structure_name)
    target_set = targets(structure)
    facet_map = facets(structure)
    asset_map = assets(structure)

    predicate, message = verify_target(target, target_set)
    if not predicate:
        log.error(f'failed verification with: {message}')
        return 1
    log.info(f'- target ({target}) OK')

    predicate, message = verify_facet(facet, target, facet_map)
    if not predicate:
        log.error(f'failed verification with: {message}')
        return 1
    log.info(f'- facet ({facet}) of target ({target}) OK')

    predicate, message = verify_assets(facet, target, asset_map)
    if not predicate:
        log.error(f'failed verification with: {message}')
        return 1
    log.info(f'- assets ({", ".join(sorted(KEYS_REQUIRED))}) for facet ({facet}) of target ({target}) OK')

    signatures_path = asset_map[target][facet][KEY_APPROVALS]
    log.info(f'loading signatures from {signatures_path=}')
    signatures = load_approvals(facet, target, signatures_path)
    log.info(f'{signatures=}')
    history_path = asset_map[target][facet][KEY_CHANGES]
    log.info(f'loading history from {history_path=}')
    history = load_changes(facet, target, history_path)
    log.info(f'{history=}')

    layout_path = asset_map[target][facet][KEY_LAYOUT]
    log.info(f'loading layout from {layout_path=}')
    design = load_layout(facet, target, layout_path)
    log.info(f'{design=}')

    metadata_path = asset_map[target][facet][KEY_META]
    log.info(f'loading metadata from {metadata_path=}')
    info = load_meta(facet, target, metadata_path)

    log.info(f'{info=}')
    log.info('successful verification')
    return 0



"""Eject templates and configurations."""

import pathlib

import liitos.template as tpl
from liitos import ENCODING, log

THINGS = {
    'bookmatter-pdf': (BOOKMATTER_TEMPLATE := 'templates/bookmatter.tex.in'),
    'driver-pdf': (DRIVER_TEMPLATE := 'templates/driver.tex.in'),
    'metadata-pdf': (METADATA_TEMPLATE := 'templates/metadata.tex.in'),
    'publisher-pdf': (PUBLISHER_TEMPLATE := 'templates/publisher.tex.in'),
    'setup-pdf': (SETUP_TEMPLATE := 'templates/setup.tex.in'),
    'approvals-yaml': (APPROVALS_YAML := 'templates/approvals.yml'),
    'changes-yaml': (CHANGES_YAML := 'templates/changes.yml'),
    'layout-yaml': (LAYOUT_YAML := 'templates/layout.yml'),
    'meta-base-yaml': (META_YAML := 'templates/meta.yml'),
    'meta-patch-yaml': (META_PATCH_YAML := 'templates/meta-patch.yml'),
    'mkdocs-yaml': (MKDOCS_PATCH_YAML := 'templates/mkdocs.yml.in'),
    'vocabulary-yaml': (VOCABULARY_YAML := 'templates/vocabulary.yml'),
}


def this(thing: str, out: str = '') -> int:
    """Later Alligator."""
    if not thing:
        log.error('eject of template with no name requested')
        log.info(f'templates known: ({", ".join(sorted(THINGS))})')
        return 2
    guesses = sorted(entry for entry in THINGS if entry.startswith(thing))
    if not guesses:
        log.error(f'eject of unknown template ({thing}) requested')
        log.info(f'templates known: ({", ".join(sorted(THINGS))})')
        return 2
    if len(guesses) > 1:
        log.error(f'eject of ambiguous template ({thing}) requested - matches ({", ".join(guesses)})')
        return 2
    content = tpl.load_resource(THINGS[guesses[0]], False)
    if not out:
        print(content)
        return 0

    out_path = pathlib.Path(out)
    out_name = out_path.name
    if not THINGS[guesses[0]].endswith(out_name):
        log.warning(f'requested writing ({THINGS[guesses[0]]}) to file ({out_name})')
    with open(out_path, 'wt', encoding=ENCODING) as handle:
        handle.write(content)
    return 0



import os
import pathlib
import tempfile

import liitos.template as tpl

ALIEN_PATH = pathlib.Path('test', 'fixtures', 'template', 'alien')
META_PATCH_CONTENT = """\
---
document:
  import: meta.yml
  patch:
    header_date: PUBLICATIONDATE
    header_id: P99999
"""


def test_load_meta_patch():
    assert tpl.load_resource('templates/meta-patch.yml', False) == META_PATCH_CONTENT


def test_load_alien():
    assert int(tpl.load_resource(ALIEN_PATH, True).strip()) == 42


def test_eject():
    with tempfile.TemporaryDirectory() as tmpdirname:
        assert tpl.eject([tmpdirname]) == 0


def test_eject_default():
    prior = os.getcwd()
    with tempfile.TemporaryDirectory() as tmpdirname:
        os.chdir(tmpdirname)
        assert tpl.eject() == 0
    os.chdir(prior)



"""Loader function for templates."""

import pathlib
import pkgutil
from typing import Union

from liitos import ENCODING, PathLike

RESOURCES = (
    'templates/approvals.yml',
    'templates/bookmatter.tex.in',
    'templates/changes.yml',
    'templates/driver.tex.in',
    'templates/layout.yml',
    'templates/meta.yml',
    'templates/meta-patch.yml',
    'templates/metadata.tex.in',
    'templates/mkdocs.yml.in',
    'templates/publisher.tex.in',
    'templates/setup.tex.in',
    'templates/vocabulary.yml',
)


def load_resource(resource: PathLike, is_complete_path: bool = False) -> str:
    """Load the template either from the package resources or an external path."""
    if is_complete_path:
        with open(resource, 'rt', encoding=ENCODING) as handle:
            return handle.read()
    else:
        return pkgutil.get_data(__package__, str(resource)).decode(encoding=ENCODING)  # type: ignore


def eject(argv: Union[list[str], None] = None) -> int:
    """Eject the templates into the folder given (default EJECTED) and create the folder if it does not exist."""
    argv = argv if argv else ['']
    into = argv[0]
    if not into.strip():
        into = 'EJECTED'
    into_path = pathlib.Path(into)
    (into_path / 'templates').mkdir(parents=True, exist_ok=True)
    for resource in RESOURCES:
        write_to = into_path / resource
        data = pkgutil.get_data(__package__, resource).decode(encoding=ENCODING)  # type: ignore
        with open(write_to, 'wt', encoding=ENCODING) as target:
            target.write(data)

    return 0



"""Weave the content of the approvals data file into the output structure (for now LaTeX).

# Supported Table Layouts

# Layout `south` (the old default)

| Approvals  | Name         | Date and Signature |
|:-----------|:-------------|:-------------------|
| Author     | Au. Thor.    |                    |
| Reviewer   | Re. Viewer.  |                    |
| Approver   | Ap. Prover.  |                    |
| Authorizer | Au. Thorizer |                    |

Table: The table is simple to grow by appending rows.

The southern layout relies on the skeleton in the bookmatter.tex.in template.

# Layout `east` (the new default)

| Department          |    AB    |     AB     |     AB     |     AB.      |
|:--------------------|:--------:|:----------:|:----------:|:------------:|
| Approvals           |  Author  |  Reviewer  |  Approver  |  Authorizer  |
| Name                | Au. Thor | Re. Viewer | Ap. Prover | Au. Thorizer |
| Date <br> Signature |          |            |            |              |

Table: This table can only grow towards the right margin and a limit of only 4 role bearers is reasonable

The eastern layout requires more dynamic LaTeX generation and thus generates the construct
from the data inside this module.

For more than 4 role bearers a second table should be placed below the first, to keep the cell content readable.
"""

import os
import pathlib
from typing import Union, no_type_check

import liitos.gather as gat
import liitos.template as tpl
import liitos.tools as too
from liitos import ENCODING, KNOWN_APPROVALS_STRATEGIES, LOG_SEPARATOR, PathLike, log

BOOKMATTER_TEMPLATE = os.getenv('LIITOS_BOOKMATTER_TEMPLATE', '')
BOOKMATTER_TEMPLATE_IS_EXTERNAL = bool(BOOKMATTER_TEMPLATE)
if not BOOKMATTER_TEMPLATE:
    BOOKMATTER_TEMPLATE = 'templates/bookmatter.tex.in'

BOOKMATTER_PATH = pathlib.Path('render/pdf/bookmatter.tex')
TOKEN_EXTRA_PUSHDOWN = r'\ExtraPushdown'  # nosec B105
EXTRA_OFFSET_EM = 24
TOKEN = r'\ \mbox{THE.ROLE.SLOT} & \mbox{THE.NAME.SLOT} & \mbox{} \\[0.5ex]'  # nosec B105
ROW_TEMPLATE = r'\ \mbox{role} & \mbox{name} & \mbox{} \\[0.5ex]'
GLUE = '\n\\hline\n'
FORMAT_DATE = '%d %b %Y'
JSON_CHANNEL = 'json'
YAML_CHANNEL = 'yaml'
COLUMNS_EXPECTED = ['name', 'role', 'orga']
APPROVALS_CUT_MARKER_TOP = '% |-- approvals - cut - marker - top -->'
APPROVALS_CUT_MARKER_BOTTOM = '% <-- approvals - cut - marker - bottom --|'

LAYOUT_SOUTH_CUT_MARKER_TOP = '% |-- layout south - cut - marker - top -->'
LAYOUT_SOUTH_CUT_MARKER_BOTTOM = '% <-- layout south - cut - marker - bottom --|'

EASTERN_TABLE_MAX_MEMBERS = 4
EASTERN_TOTAL_MAX_MEMBERS = EASTERN_TABLE_MAX_MEMBERS * 2

NL = '\n'
BASE_TABLE = r"""% |-- layout east - cut - marker - top -->
\begin{large}
\addtolength\aboverulesep{0.15ex}  % extra spacing above and below rules
\addtolength\belowrulesep{0.35ex}
\begin{longtable}[]{|
 >{\raggedright\arraybackslash}m{(\columnwidth - 12\tabcolsep) * \real{0.2000}}|% <- fixed
$HEAD.BLOCK$}
\hline
\begin{minipage}[b]{\linewidth}\raggedright\ \mbox{\textbf{\theApprovalsDepartmentLabel}}\end{minipage}%
$ORGA.BLOCK$ \\[0.5ex]
\hline
\ \mbox{\textbf{\theApprovalsRoleLabel}}%
$ROLE.BLOCK$
 \\[0.5ex]
\hline
\ \mbox{\textbf{\theApprovalsNameLabel}}%
$NAME.BLOCK$
 \\[0.5ex]
\hline
\ \mbox{\textbf{Date}} \mbox{\textbf{\ }} \mbox{\textbf{\ Signature}}%
$SIGN.BLOCK$
 \\[0.5ex]
\hline

\end{longtable}
\end{large}
% <-- layout east - cut - marker - bottom --|
"""

HEAD_CELL = r' >{\raggedright\arraybackslash}m{(\columnwidth - 12\tabcolsep) * \real{0.2000}}|'
ORGA_CELL = r' & \begin{minipage}[b]{\linewidth}\raggedright\mbox{\textbf{THE.ORGA$RANK$.SLOT}}\end{minipage}'
ROLE_CELL = r' & \mbox{THE.ROLE$RANK$.SLOT}'
NAME_CELL = r' & \mbox{THE.NAME$RANK$.SLOT}'
SIGN_CELL = r' & \mbox{}'


def eastern_scaffold(normalized: list[dict[str, str]]) -> str:
    """Inject the blocks derived from the approvals data to yield the fill-in scaffold."""
    bearers = len(normalized)
    table_max_members = EASTERN_TABLE_MAX_MEMBERS
    total_max_members = EASTERN_TOTAL_MAX_MEMBERS
    if bearers > total_max_members:
        raise NotImplementedError(
            f'Please use southwards layout for more than {total_max_members} role bearers;'
            f' found ({bearers}) entries in approvals data source.'
        )

    # First up to 4 entries got into upper table and final upt to 4 entries (if any) to lower table
    upper, lower = normalized[:table_max_members], normalized[table_max_members:]
    uppers, lowers = len(upper), len(lower)
    log.info(f'SPLIT {uppers}, {lowers}, {bearers}')
    log.info(f'UPPER: {list(range(uppers))}')
    head_block = (f'{HEAD_CELL}{NL}' * uppers).rstrip(NL)
    orga_block = NL.join(ORGA_CELL.replace('$RANK$', str(slot)) for slot in range(uppers))
    role_block = NL.join(ROLE_CELL.replace('$RANK$', str(slot)) for slot in range(uppers))
    name_block = NL.join(NAME_CELL.replace('$RANK$', str(slot)) for slot in range(uppers))
    sign_block = f'{SIGN_CELL}{NL}' * uppers
    upper_table = (
        BASE_TABLE.replace('$HEAD.BLOCK$', head_block)
        .replace('$ORGA.BLOCK$', orga_block)
        .replace('$ROLE.BLOCK$', role_block)
        .replace('$NAME.BLOCK$', name_block)
        .replace('$SIGN.BLOCK$', sign_block)
    )

    for thing in upper_table.split(NL):
        log.debug(thing)

    if not lowers:
        return upper_table

    log.info(f'LOWER: {list(range(uppers, bearers))}')
    head_block = (f'{HEAD_CELL}{NL}' * lowers).rstrip(NL)
    orga_block = NL.join(ORGA_CELL.replace('$RANK$', str(slot)) for slot in range(uppers, bearers))
    role_block = NL.join(ROLE_CELL.replace('$RANK$', str(slot)) for slot in range(uppers, bearers))
    name_block = NL.join(NAME_CELL.replace('$RANK$', str(slot)) for slot in range(uppers, bearers))
    sign_block = f'{SIGN_CELL}{NL}' * lowers

    lower_table = (
        BASE_TABLE.replace('$HEAD.BLOCK$', head_block)
        .replace('$ORGA.BLOCK$', orga_block)
        .replace('$ROLE.BLOCK$', role_block)
        .replace('$NAME.BLOCK$', name_block)
        .replace('$SIGN.BLOCK$', sign_block)
    )

    for thing in lower_table.split(NL):
        log.debug(thing)

    return f'{upper_table}{NL}{lower_table}'


def get_layout(layout_path: PathLike, target_key: str, facet_key: str) -> dict[str, dict[str, dict[str, bool]]]:
    """Boolean layout decisions on bookmatter and publisher page conten.

    Deprecated as the known use cases evolved into a different direction ...
    """
    layout = {'layout': {'global': {'has_approvals': True, 'has_changes': True, 'has_notices': True}}}
    if layout_path:
        log.info(f'loading layout from {layout_path=} for approvals')
        return gat.load_layout(facet_key, target_key, layout_path)[0]  # type: ignore

    log.info('using default layout for approvals')
    return layout


def derive_model(model_path: PathLike) -> tuple[str, list[str]]:
    """Derive the model as channel type and column model from the given path."""
    channel = JSON_CHANNEL if str(model_path).endswith('.json') else YAML_CHANNEL
    columns_expected = ['Approvals', 'Name'] if channel == JSON_CHANNEL else COLUMNS_EXPECTED

    return channel, columns_expected


def columns_are_present(columns_present: list[str], columns_expected: list[str]) -> bool:
    """Ensure the needed columns are present."""
    return all(column in columns_expected for column in columns_present)


@no_type_check
def normalize(signatures: object, channel: str, columns_expected: list[str]) -> list[dict[str, str]]:
    """Normalize the channel specific topology of the model into a logical model.

    On error an empty logical model is returned.
    """
    if channel == JSON_CHANNEL:
        if not columns_are_present(signatures[0]['columns'], columns_expected):
            log.error('unexpected column model!')
            log.error(f'-  expected: ({columns_expected})')
            log.error(f'- but found: ({signatures[0]["columns"]})')
            return []

    if channel == YAML_CHANNEL:
        for slot, approval in enumerate(signatures[0]['approvals'], start=1):
            log.debug(f'{slot=}, {approval=}')
            if not columns_are_present(approval, columns_expected):
                log.error('unexpected column model!')
                log.error(f'-  expected: ({columns_expected})')
                log.error(f'- but found: ({sorted(approval)}) in slot #{slot}')
                return []

    if channel == JSON_CHANNEL:
        return [{'role': role, 'name': name} for role, name in signatures[0]['rows']]

    return [
        {'orga': approval.get('orga', ''), 'role': approval['role'], 'name': approval['name']}
        for approval in signatures[0]['approvals']
    ]


def inject_southwards(lines: list[str], rows: list[str], pushdown: float) -> None:
    """Deploy approvals data per southern layout strategy per updating the lines list in place."""
    for n, line in enumerate(lines):
        if TOKEN_EXTRA_PUSHDOWN in line:
            lines[n] = line.replace(TOKEN_EXTRA_PUSHDOWN, f'{pushdown}em')
            continue
        if line == TOKEN:
            lines[n] = GLUE.join(rows)
            break
    if lines[-1]:  # Need separating empty line?
        lines.append(NL)


def inject_eastwards(lines: list[str], normalized: list[dict[str, str]], pushdown: float) -> None:
    """Deploy approvals data per eastern layout strategy per updating the lines list in place."""
    for n, line in enumerate(lines):
        if TOKEN_EXTRA_PUSHDOWN in line:
            lines[n] = line.replace(TOKEN_EXTRA_PUSHDOWN, f'{pushdown}em')
            break
    hack = eastern_scaffold(normalized)
    log.info('logical model for approvals table is:')
    for slot, entry in enumerate(normalized):
        orga = entry['orga'] if entry['orga'] else r'\theApprovalsDepartmentValue'
        log.info(f'- {entry["role"]} <-- {entry["name"]} (from {orga})')
        hack = (
            hack.replace(f'THE.ORGA{slot}.SLOT', orga)
            .replace(f'THE.ROLE{slot}.SLOT', entry['role'])
            .replace(f'THE.NAME{slot}.SLOT', entry['name'])
        )
    lines.extend(hack.split(NL))
    if lines[-1]:  # pragma: no cover
        lines.append(NL)


def weave(
    doc_root: Union[str, pathlib.Path],
    structure_name: str,
    target_key: str,
    facet_key: str,
    options: dict[str, Union[bool, str]],
) -> int:
    """Map the approvals data to a table on the titlepage."""
    log.info(LOG_SEPARATOR)
    log.info('entered signatures weave function ...')
    structure, asset_map = gat.prelude(
        doc_root=doc_root,
        structure_name=structure_name,
        target_key=target_key,
        facet_key=facet_key,
        command='approvals',
    )

    layout_path = asset_map[target_key][facet_key].get(gat.KEY_LAYOUT, '')
    layout = get_layout(layout_path, target_key=target_key, facet_key=facet_key)
    log.info(f'{layout=}')

    log.info(LOG_SEPARATOR)
    signatures_path = asset_map[target_key][facet_key][gat.KEY_APPROVALS]
    channel, columns_expected = derive_model(signatures_path)
    log.info(f'detected approvals channel ({channel}) weaving in from ({signatures_path})')

    log.info(f'loading signatures from {signatures_path=}')
    signatures = gat.load_approvals(facet_key, target_key, signatures_path)
    log.info(f'{signatures=}')

    log.info(LOG_SEPARATOR)
    log.info('plausibility tests for approvals ...')

    logical_model = normalize(signatures, channel=channel, columns_expected=columns_expected)

    rows = [ROW_TEMPLATE.replace('role', kv['role']).replace('name', kv['name']) for kv in logical_model]

    pushdown = EXTRA_OFFSET_EM - 2 * len(rows)
    log.info(f'calculated extra pushdown to be {pushdown}em')

    bookmatter_template = tpl.load_resource(BOOKMATTER_TEMPLATE, BOOKMATTER_TEMPLATE_IS_EXTERNAL)
    lines = [line.rstrip() for line in bookmatter_template.split('\n')]

    if not layout['layout']['global']['has_approvals']:
        log.info('removing approvals from document layout')
        lines = list(too.remove_target_region_gen(lines, APPROVALS_CUT_MARKER_TOP, APPROVALS_CUT_MARKER_BOTTOM))

    log.info(LOG_SEPARATOR)
    log.info(f'weaving in the approvals from {signatures_path}...')
    approvals_strategy = options.get('approvals_strategy', KNOWN_APPROVALS_STRATEGIES[0])
    log.info(f'selected approvals layout strategy is ({approvals_strategy})')
    if approvals_strategy == 'south':
        inject_southwards(lines, rows, pushdown)
    else:  # default is east
        lines = list(too.remove_target_region_gen(lines, LAYOUT_SOUTH_CUT_MARKER_TOP, LAYOUT_SOUTH_CUT_MARKER_BOTTOM))
        inject_eastwards(lines, logical_model, pushdown)

    effective_path = pathlib.Path(layout_path).parent / BOOKMATTER_PATH
    log.info(f'Writing effective bookmatter file to ({effective_path})')
    with open(effective_path, 'wt', encoding=ENCODING) as handle:
        handle.write('\n'.join(lines))
    log.info(LOG_SEPARATOR)

    return 0



"""Weave the content of the changes data file into the output structure (for now LaTeX)."""

import os
import pathlib
from typing import Generator, Union, no_type_check

import liitos.gather as gat
import liitos.template as tpl
import liitos.tools as too
from liitos import ENCODING, LOG_SEPARATOR, PathLike, log


PUBLISHER_TEMPLATE = os.getenv('LIITOS_PUBLISHER_TEMPLATE', '')
PUBLISHER_TEMPLATE_IS_EXTERNAL = bool(PUBLISHER_TEMPLATE)
if not PUBLISHER_TEMPLATE:
    PUBLISHER_TEMPLATE = 'templates/publisher.tex.in'

PUBLISHER_PATH = pathlib.Path('render/pdf/publisher.tex')
TOKEN = r'THE.ISSUE.CODE & THE.REVISION.CODE & THE.AUTHOR.NAME & THE.DESCRIPTION \\'  # nosec B105
DEFAULT_REVISION = '00'
ROW_TEMPLATE = r'issue & revision & author & summary \\'
GLUE = '\n\\hline\n'
JSON_CHANNEL = 'json'
YAML_CHANNEL = 'yaml'
COLUMNS_EXPECTED = sorted(['author', 'date', 'issue', 'revision', 'summary'])
COLUMNS_MINIMAL = sorted(['author', 'issue', 'summary'])
CUT_MARKER_CHANGES_TOP = '% |-- changes - cut - marker - top -->'
CUT_MARKER_CHANGES_BOTTOM = '% <-- changes - cut - marker - bottom --|'
CUT_MARKER_NOTICES_TOP = '% |-- notices - cut - marker - top -->'
CUT_MARKER_NOTICES_BOTTOM = '% <-- notices - cut - marker - bottom --|'
TOKEN_ADJUSTED_PUSHDOWN = r'\AdustedPushdown'  # nosec B105
DEFAULT_ADJUSTED_PUSHDOWN_VALUE = 14

NL = '\n'


def get_layout(layout_path: PathLike, target_key: str, facet_key: str) -> dict[str, dict[str, dict[str, bool]]]:
    """Boolean layout decisions on bookmatter and publisher page conten.

    Deprecated as the known use cases evolved into a different direction ...
    """
    layout = {'layout': {'global': {'has_approvals': True, 'has_changes': True, 'has_notices': True}}}
    if layout_path:
        log.info(f'loading layout from {layout_path=} for changes and notices')
        return gat.load_layout(facet_key, target_key, layout_path)[0]  # type: ignore

    log.info('using default layout for approvals')
    return layout


def derive_model(model_path: PathLike) -> tuple[str, list[str]]:
    """Derive the model as channel type and column model from the given path."""
    channel = JSON_CHANNEL if str(model_path).endswith('.json') else YAML_CHANNEL
    columns_expected = COLUMNS_MINIMAL if channel == JSON_CHANNEL else COLUMNS_EXPECTED

    return channel, columns_expected


def columns_are_present(columns_present: list[str], columns_expected: list[str]) -> bool:
    """Ensure the needed columns are present."""
    return all(column in columns_expected for column in columns_present)


@no_type_check
def normalize(changes: object, channel: str, columns_expected: list[str]) -> list[dict[str, str]]:
    """Normalize the channel specific topology of the model into a logical model.

    On error an empty logical model is returned.
    """
    if channel == JSON_CHANNEL:
        for slot, change in enumerate(changes[0]['changes'], start=1):
            if not set(columns_expected).issubset(set(change)):
                log.error('unexpected column model!')
                log.error(f'-  expected: ({columns_expected})')
                log.error(f'-   minimal: ({COLUMNS_MINIMAL})')
                log.error(f'- but found: ({change}) for entry #{slot}')
                return []

    if channel == YAML_CHANNEL:
        for slot, change in enumerate(changes[0]['changes'], start=1):
            model = sorted(change.keys())
            if not set(COLUMNS_MINIMAL).issubset(set(model)):
                log.error('unexpected column model!')
                log.error(f'-  expected: ({columns_expected})')
                log.error(f'-   minimal: ({COLUMNS_MINIMAL})')
                log.error(f'- but found: ({model}) in slot {slot}')
                return []

    model = []
    if channel == JSON_CHANNEL:
        for change in changes[0]['changes']:
            issue, author, summary = change['issue'], change['author'], change['summary']
            revision = change.get('revision', DEFAULT_REVISION)
            model.append({'issue': issue, 'revision': revision, 'author': author, 'summary': summary})
        return model

    for change in changes[0]['changes']:
        author = change['author']
        issue = change['issue']
        revision = change.get('revision', DEFAULT_REVISION)
        summary = change['summary']
        model.append({'issue': issue, 'revision': revision, 'author': author, 'summary': summary})

    return model


def adjust_pushdown_gen(text_lines: list[str], pushdown: float) -> Generator[str, None, None]:
    """Update the pushdown line filtering the incoming lines."""
    for line in text_lines:
        if TOKEN_ADJUSTED_PUSHDOWN in line:
            line = line.replace(TOKEN_ADJUSTED_PUSHDOWN, f'{pushdown}em')
            log.info(f'set adjusted pushdown value {pushdown}em')
        yield line


def weave(
    doc_root: Union[str, pathlib.Path],
    structure_name: str,
    target_key: str,
    facet_key: str,
    options: dict[str, Union[bool, str]],
) -> int:
    """Later alligator."""
    log.info(LOG_SEPARATOR)
    log.info('entered changes weave function ...')
    structure, asset_map = gat.prelude(
        doc_root=doc_root, structure_name=structure_name, target_key=target_key, facet_key=facet_key, command='changes'
    )

    layout_path = asset_map[target_key][facet_key].get(gat.KEY_LAYOUT, '')
    layout = get_layout(layout_path, target_key=target_key, facet_key=facet_key)
    log.info(f'{layout=}')

    log.info(LOG_SEPARATOR)
    changes_path = asset_map[target_key][facet_key][gat.KEY_CHANGES]
    channel, columns_expected = derive_model(changes_path)
    log.info(f'detected changes channel ({channel}) weaving in from ({changes_path})')

    log.info(f'loading changes from {changes_path=}')
    changes = gat.load_changes(facet_key, target_key, changes_path)
    log.info(f'{changes=}')

    log.info(LOG_SEPARATOR)
    log.info('plausibility tests for changes ...')

    logical_model = normalize(changes, channel=channel, columns_expected=columns_expected)

    rows = [
        ROW_TEMPLATE.replace('issue', kv['issue'])
        .replace('revision', kv['revision'])
        .replace('author', kv['author'])
        .replace('summary', kv['summary'])
        for kv in logical_model
    ]

    pushdown = DEFAULT_ADJUSTED_PUSHDOWN_VALUE
    log.info(f'calculated adjusted pushdown to be {pushdown}em')

    publisher_template = tpl.load_resource(PUBLISHER_TEMPLATE, PUBLISHER_TEMPLATE_IS_EXTERNAL)
    lines = [line.rstrip() for line in publisher_template.split(NL)]

    if any(TOKEN_ADJUSTED_PUSHDOWN in line for line in lines):
        lines = list(adjust_pushdown_gen(lines, pushdown))
    else:
        log.error(f'token ({TOKEN_ADJUSTED_PUSHDOWN}) not found - template mismatch')

    if not layout['layout']['global']['has_changes']:
        log.info('removing changes from document layout')
        lines = list(too.remove_target_region_gen(lines, CUT_MARKER_CHANGES_TOP, CUT_MARKER_CHANGES_BOTTOM))

    if not layout['layout']['global']['has_notices']:
        log.info('removing notices from document layout')
        lines = list(too.remove_target_region_gen(lines, CUT_MARKER_NOTICES_TOP, CUT_MARKER_NOTICES_BOTTOM))

    log.info(LOG_SEPARATOR)
    log.info('weaving in the changes from {changes_path} ...')
    for n, line in enumerate(lines):
        if line.strip() == TOKEN:
            lines[n] = GLUE.join(rows)
            break
    if lines[-1]:
        lines.append('\n')
    with open(PUBLISHER_PATH, 'wt', encoding=ENCODING) as handle:
        handle.write('\n'.join(lines))
    log.info(LOG_SEPARATOR)

    return 0



import copy
import os
import pathlib

import pytest

import liitos.gather as gather
from liitos import KEY_LAYOUT

TEST_PREFIX = pathlib.Path('test', 'fixtures', 'basic')
DEFAULT_STRUCTURE_PATH = TEST_PREFIX / gather.DEFAULT_STRUCTURE_NAME
TEST_TARGET = 'abc'
TEST_FACET = 'mn'
TEST_STRUCTURE = {
    TEST_TARGET: {
        TEST_FACET: {
            gather.KEY_APPROVALS: 'approvals.json',
            gather.KEY_BIND: 'bind-mn.txt',
            gather.KEY_CHANGES: 'changes.json',
            gather.KEY_META: 'meta-mn.yml',
        }
    }
}
TEST_MAKE_MISSING = 'missing-this-file-for-'

LAYOUT_ALL_PATH = pathlib.Path('test/fixtures/layout/all.yml')


def test_load_structure():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assert TEST_TARGET in structure
    assert TEST_STRUCTURE[TEST_TARGET] in structure[TEST_TARGET]


def test_targets():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    targets = gather.targets(structure)
    assert targets == set([TEST_TARGET])


def test_facets():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    facets = gather.facets(structure)
    assert facets == {TEST_TARGET: {'missing', 'opq', TEST_FACET}}


def test_assets():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    assert assets[TEST_TARGET][TEST_FACET] == TEST_STRUCTURE[TEST_TARGET][TEST_FACET]


def test_changes():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    assets[TEST_TARGET][TEST_FACET][gather.KEY_CHANGES] = str(TEST_PREFIX / 'changes.json')
    changes, message = gather.changes(TEST_FACET, TEST_TARGET, assets)
    assert not message
    assert changes == {
        'columns': ['issue', 'author', 'date', 'summary'],
        'rows': [['01', 'One Author', '31.12.2024', 'Initial Issue']],
    }


def test_changes_key_missing():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    del assets[TEST_TARGET][TEST_FACET][gather.KEY_CHANGES]
    changes, message = gather.changes(TEST_FACET, TEST_TARGET, assets)
    assert message
    assert not changes


def test_changes_link_missing():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    assets[TEST_TARGET][TEST_FACET][gather.KEY_CHANGES] = str(TEST_PREFIX / f'{TEST_MAKE_MISSING}changes.json')
    changes, message = gather.changes(TEST_FACET, TEST_TARGET, assets)
    assert message
    assert not changes


def test_approvals():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    assets[TEST_TARGET][TEST_FACET][gather.KEY_APPROVALS] = str(TEST_PREFIX / 'approvals.json')
    approvals, message = gather.approvals(TEST_FACET, TEST_TARGET, assets)
    assert not message
    assert approvals == {
        'columns': ['Approvals', 'Name'],
        'rows': [['Author', 'One Author'], ['Review', 'One Reviewer'], ['Approved', 'One Approver']],
    }


def test_approvals_key_missing():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    del assets[TEST_TARGET][TEST_FACET][gather.KEY_APPROVALS]
    approvals, message = gather.approvals(TEST_FACET, TEST_TARGET, assets)
    assert message
    assert not approvals


def test_approvals_link_missing():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    assets[TEST_TARGET][TEST_FACET][gather.KEY_APPROVALS] = str(TEST_PREFIX / f'{TEST_MAKE_MISSING}approvals.json')
    approvals, message = gather.approvals(TEST_FACET, TEST_TARGET, assets)
    assert message
    assert not approvals


def test_meta_key_missing():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    del assets[TEST_TARGET][TEST_FACET][gather.KEY_META]
    meta, message = gather.meta(TEST_FACET, TEST_TARGET, assets)
    assert message
    assert not meta


def test_meta_link_missing():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    assets[TEST_TARGET][TEST_FACET][gather.KEY_META] = str(TEST_PREFIX / f'{TEST_MAKE_MISSING}meta.json')
    meta, message = gather.meta(TEST_FACET, TEST_TARGET, assets)
    assert message
    assert not meta


def test_binder_key_missing():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    del assets[TEST_TARGET][TEST_FACET][gather.KEY_BIND]
    binder, message = gather.binder(TEST_FACET, TEST_TARGET, assets)
    assert message
    assert not binder


def test_binder_link_missing():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    assets[TEST_TARGET][TEST_FACET][gather.KEY_BIND] = str(TEST_PREFIX / f'{TEST_MAKE_MISSING}bind-mn.txt')
    binder, message = gather.binder(TEST_FACET, TEST_TARGET, assets)
    assert message
    assert not binder


def test_binder():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    assets[TEST_TARGET][TEST_FACET][gather.KEY_BIND] = str(TEST_PREFIX / f'bind-{TEST_FACET}.txt')
    binder, message = gather.binder(TEST_FACET, TEST_TARGET, assets)
    assert not message
    assert binder == ['head.md', 'body.md']


def test_meta():
    structure = gather.load_structure(DEFAULT_STRUCTURE_PATH)
    assets = gather.assets(structure)
    assets[TEST_TARGET][TEST_FACET][gather.KEY_META] = str(TEST_PREFIX / f'meta-{TEST_FACET}.md')
    meta, message = gather.meta(TEST_FACET, TEST_TARGET, assets)
    assert not message
    assert meta == {'setting': 'value'}


def test_verify_target():
    targets = set([TEST_TARGET])
    predicate, message = gather.verify_target(TEST_TARGET, targets)
    assert not message
    assert predicate


def test_verify_target_not():
    targets = set([TEST_TARGET])
    predicate, message = gather.verify_target(f'{TEST_MAKE_MISSING}{TEST_TARGET}', targets)
    assert message == f'target ({TEST_MAKE_MISSING}{TEST_TARGET}) not in {sorted(targets)}'
    assert not predicate


def test_verify_facet():
    facets = {TEST_TARGET: set([TEST_FACET])}
    predicate, message = gather.verify_facet(TEST_FACET, TEST_TARGET, facets)
    assert not message
    assert predicate


def test_verify_facet_not():
    facets = {TEST_TARGET: set([TEST_FACET])}
    predicate, message = gather.verify_facet(f'{TEST_MAKE_MISSING}{TEST_FACET}', TEST_TARGET, facets)
    expected = (
        f'facet ({TEST_MAKE_MISSING}{TEST_FACET})' f' of target ({TEST_TARGET}) not in {sorted(facets[TEST_TARGET])}'
    )
    assert message == expected
    assert not predicate


def test_verify_asset_keys():
    assets = copy.deepcopy(TEST_STRUCTURE)
    predicate, message = gather.verify_asset_keys(TEST_FACET, TEST_TARGET, assets)
    assert not message
    assert predicate


def test_verify_asset_keys_not():
    assets = copy.deepcopy(TEST_STRUCTURE)
    del assets[TEST_TARGET][TEST_FACET][gather.KEY_BIND]
    predicate, message = gather.verify_asset_keys(TEST_FACET, TEST_TARGET, assets)
    expected = (
        f'keys in {sorted(gather.KEYS_REQUIRED)}' f' for facet ({TEST_FACET}) of target ({TEST_TARGET}) are missing'
    )
    assert message == expected
    assert not predicate


def test_verify_asset_links():
    assets = copy.deepcopy(TEST_STRUCTURE)
    ole_place = pathlib.Path.cwd()
    os.chdir(TEST_PREFIX)
    predicate, message = gather.verify_asset_links(TEST_FACET, TEST_TARGET, assets)
    os.chdir(ole_place)
    assert not message
    assert predicate


def test_verify_asset_links_no_key():
    assets = copy.deepcopy(TEST_STRUCTURE)
    del assets[TEST_TARGET][TEST_FACET][gather.KEY_BIND]
    ole_place = pathlib.Path.cwd()
    os.chdir(TEST_PREFIX)
    predicate, message = gather.verify_asset_links(TEST_FACET, TEST_TARGET, assets)
    os.chdir(ole_place)
    expected = (
        f'keys in {sorted(gather.KEYS_REQUIRED)}' f' for facet ({TEST_FACET}) of target ({TEST_TARGET}) are missing'
    )
    assert message == expected
    assert not predicate


def test_verify_asset_links_no_link():
    assets = copy.deepcopy(TEST_STRUCTURE)
    bad_link_value = f'{TEST_MAKE_MISSING}bind-{TEST_FACET}.txt'
    assets[TEST_TARGET][TEST_FACET][gather.KEY_BIND] = bad_link_value
    ole_place = pathlib.Path.cwd()
    os.chdir(TEST_PREFIX)
    predicate, message = gather.verify_asset_links(TEST_FACET, TEST_TARGET, assets)
    os.chdir(ole_place)
    expected = f'bind asset link ({bad_link_value})' f' for facet ({TEST_FACET}) of target ({TEST_TARGET}) is invalid'
    assert message == expected
    assert not predicate


def test_verify_assets():
    assets = copy.deepcopy(TEST_STRUCTURE)
    ole_place = pathlib.Path.cwd()
    os.chdir(TEST_PREFIX)
    predicate, message = gather.verify_assets(TEST_FACET, TEST_TARGET, assets)
    os.chdir(ole_place)
    assert message == 'layout asset for facet (mn) of target (abc) is invalid'
    assert predicate is False


def test_verify_assets_no_key():
    assets = copy.deepcopy(TEST_STRUCTURE)
    del assets[TEST_TARGET][TEST_FACET][gather.KEY_BIND]
    ole_place = pathlib.Path.cwd()
    os.chdir(TEST_PREFIX)
    predicate, message = gather.verify_assets(TEST_FACET, TEST_TARGET, assets)
    os.chdir(ole_place)
    expected = (
        f'keys in {sorted(gather.KEYS_REQUIRED)}' f' for facet ({TEST_FACET}) of target ({TEST_TARGET}) are missing'
    )
    assert message == expected
    assert not predicate


def test_verify_assets_no_link():
    assets = copy.deepcopy(TEST_STRUCTURE)
    bad_link_value = f'{TEST_MAKE_MISSING}bind-{TEST_FACET}.txt'
    assets[TEST_TARGET][TEST_FACET][gather.KEY_BIND] = bad_link_value
    ole_place = pathlib.Path.cwd()
    os.chdir(TEST_PREFIX)
    predicate, message = gather.verify_assets(TEST_FACET, TEST_TARGET, assets)
    os.chdir(ole_place)
    expected = f'bind asset link ({bad_link_value}) for facet ({TEST_FACET}) of target ({TEST_TARGET}) is invalid'
    assert message == expected
    assert not predicate


def test_verify_assets_empty_changes():
    assets = copy.deepcopy(TEST_STRUCTURE)
    link_value_to_bad_content = 'no-changes.json'
    assets[TEST_TARGET][TEST_FACET][gather.KEY_CHANGES] = link_value_to_bad_content
    ole_place = pathlib.Path.cwd()
    os.chdir(TEST_PREFIX)
    predicate, message = gather.verify_assets(TEST_FACET, TEST_TARGET, assets)
    os.chdir(ole_place)
    expected = f'changes asset for facet ({TEST_FACET}) of target ({TEST_TARGET}) is invalid'
    assert message == expected
    assert not predicate


def test_load_layout():
    data, msg = gather.load_layout('facet', 'target', LAYOUT_ALL_PATH)
    assert data == {'layout': {'global': {'has_approvals': True, 'has_changes': True, 'has_notices': True}}}
    assert not msg


def test_load_layout_failing_path():
    facet = 'facet'
    target = 'target'
    path = 'no-file-here'
    label = 'Metadata'
    data, msg = gather.load_layout(facet, target, path)
    assert data == {}
    assert msg == f'{label} link not found at ({path}) or invalid for facet ({facet}) of target ({target})'


def test_layout():
    facet = 'facet'
    target = 'target'
    path = LAYOUT_ALL_PATH
    asset_struct = {target: {facet: {KEY_LAYOUT: path}}}
    data, msg = gather.layout(facet, target, asset_struct)
    assert data == {'layout': {'global': {'has_approvals': True, 'has_changes': True, 'has_notices': True}}}
    assert not msg


def test_layout_failing_path():
    facet = 'facet'
    target = 'target'
    path = 'no-file-here'
    asset_struct = {target: {facet: {KEY_LAYOUT: path}}}
    label = 'Metadata'
    data, msg = gather.layout(facet, target, asset_struct)
    assert data == {}
    assert msg == f'{label} link not found at ({path}) or invalid for facet ({facet}) of target ({target})'


def test_load_changes_failing_format():
    facet = 'facet'
    target = 'target'
    path = 'not-present.wrong-suffix'
    label = 'Changes'
    data, msg = gather.load_changes(facet, target, path)
    assert data == []
    assert msg == f'{label} requires json or yaml format in assets for facet ({facet}) of target ({target})'


def test_error_context_not_implemented():
    err = ZeroDivisionError()
    with pytest.raises(NotImplementedError) as excinfo:
        gather.error_context(None, 'Rookie', None, None, None, ZeroDivisionError())
    assert f'error context not implemented for error ({err})' in str(excinfo.value)



import liitos.tables as tables

TABLE_THREE_COLS = r"""\
\begin{longtable}[]{@{}lcr@{}}
\caption{A caption for a table
\label{table:left-middle-right}}\tabularnewline
\toprule()
Left & Middle & Right \\
\midrule()
\endfirsthead
\toprule()
Left & Middle & Right \\
\midrule()
\endhead
L1 & M2 & R3 \\
L4 & M5 & R6 \\
L7 & M8 & R9 \\
L10 & M11 & R12 \\
L13 & M14 & R15 \\
\bottomrule()
\end{longtable}
"""


def test_patch_empty():
    assert tables.patch([]) == []


def test_patch_tab_start():
    incoming = [
        '',
        tables.TAB_START_TOK,
        tables.TOP_RULE,
        tables.MID_RULE,
        tables.END_HEAD,
        tables.END_DATA_ROW,
        tables.BOT_RULE,
        tables.TAB_END_TOK,
    ]
    assert tables.patch(incoming) == incoming


def test_patch_tab_three_cols():
    incoming = TABLE_THREE_COLS.split()
    assert tables.patch(incoming) == incoming


def test_table():
    with open('test/fixtures/random/tables.tex', 'rt', encoding='utf-8') as handle:
        lines_buffer = [line.rstrip() for line in handle.readlines()]

    reader = iter(lines_buffer)
    some_tables = []
    comment_outs = []
    n = 0
    widths = []
    for line in reader:
        if not line.startswith(tables.Table.LBP_STARTSWITH_TAB_ENV_BEGIN):
            if line.startswith(r'\columns='):
                has_column, text_line, widths = tables.parse_columns_command(n, line)
                if has_column:
                    comment_outs.append(n)
            n += 1
        else:
            table = tables.Table(n, line, reader, widths)  # sharing the meal - instead of iter(lines_buffer[n:]))
            widths = []
            some_tables.append(table)
            n += len(some_tables[-1].source_map())

    assert comment_outs == [92]


def test_patch_some():
    with open('test/fixtures/random/tables.tex', 'rt', encoding='utf-8') as handle:
        lines_buffer = [line.rstrip() for line in handle.readlines()]
    out_lines = tables.patch(lines_buffer)
    assert out_lines[92] == r'%CONSIDERED_\columns=,10\%,30\%,50\%'


def test_parse_table_font_size_command_unknown():
    line = '\\tablefontsize=unknown'
    worked, out_line, font_size = tables.parse_table_font_size_command(0, line)
    assert worked is False
    assert out_line == line
    assert font_size == ''


def test_parse_table_font_size_command_no_command_at_start():
    line = ' \\tablefontsize=notatstart'
    worked, out_line, font_size = tables.parse_table_font_size_command(0, line)
    assert worked is False
    assert out_line == line
    assert font_size == ''


def test_parse_table_font_size_command_known_size():
    line = '\\tablefontsize=footnotesize'
    worked, out_line, font_size = tables.parse_table_font_size_command(0, line)
    assert worked is True
    assert out_line == ''
    assert font_size == 'footnotesize'


def test_parse_table_font_size_command_known_size_with_backslash():
    line = '\\tablefontsize=\\tiny'
    worked, out_line, font_size = tables.parse_table_font_size_command(0, line)
    assert worked is True
    assert out_line == ''
    assert font_size == 'tiny'


def test_parse_table_font_size_command_no_equal_sign():
    line = '\\tablefontsize\\tiny'
    worked, out_line, font_size = tables.parse_table_font_size_command(0, line)
    assert worked is False
    assert out_line == line
    assert font_size == ''


def test_parse_table_font_size_command_missing_value():
    line = '\\tablefontsize='
    worked, out_line, font_size = tables.parse_table_font_size_command(0, line)
    assert worked is False
    assert out_line == line
    assert font_size == ''



import liitos.patch as patch


def test_apply_empty_on_empty_ok():
    assert patch.apply([], []) == []


def test_apply_empty_on_some_ok():
    some = ['a', '', 'z']
    assert patch.apply([], some) == some


def test_apply_pair_on_some_ok():
    pair = [('a', 'b')]
    some = ['a', '', 'z']
    assert patch.apply(pair, some) == ['b', '', 'z']


def test_apply_pairs_on_some_ok():
    pairs = [('a', 'b'), ['a', 'b']]
    some = ['a', '', 'z']
    assert patch.apply(pairs, some) == ['b', '', 'z']


def test_apply_pairs_on_some_orderly_ok():
    pairs = [['a', 'b'], ['b', 'a']]
    some = ['a', '', 'z']
    assert patch.apply(pairs, some) == some  # type: ignore



from collections.abc import Iterable
from typing import Union

from liitos import log

NO_OPTION: str = ''


def parse_options_command(slot: int, text_line: str) -> tuple[bool, str, str]:
    """Parse the \\option[style=multiline,leftmargin=6em]."""
    if text_line.startswith(r'\option['):
        log.info(f'trigger an option mod for the next description environment at line #{slot + 1}|{text_line}')
        try:
            # \option[style=multiline,leftmargin=6em]  --> [style=multiline,leftmargin=6em]
            opt = text_line.split(r'\option', 1)[1].strip()
            log.info(f' -> parsed option as ({opt})')
            return True, f'%CONSIDERED_{text_line}', opt
        except Exception as err:
            log.error(f'failed to parse option value from {text_line.strip()} with err: {err}')
            return False, text_line, ''
    else:
        return False, text_line, ''


def options(incoming: Iterable[str], lookup: Union[dict[str, str], None] = None) -> list[str]:
    """Later alligator. \\option[style=multiline,leftmargin=6em]"""
    outgoing = []
    modus = 'copy'
    opt = NO_OPTION
    for slot, line in enumerate(incoming):
        if modus == 'copy':
            has_opt, text_line, opt = parse_options_command(slot, line)
            if has_opt:
                modus = 'option'
            else:
                outgoing.append(text_line)
            continue

        # if modus == 'option':
        if line.startswith(r'\begin{description}'):
            if opt != NO_OPTION:
                log.info(f'- found the option target start at line #{slot + 1}|{line}')
                outgoing.append(f'\\begin{{description}}{opt}')
            else:
                outgoing.append(line)
            modus = 'copy'
            opt = NO_OPTION
        else:
            outgoing.append(line)

    return outgoing



"""Given a target and facet, concatenate a tree of markdown files to a single file rewriting all image refs."""

import json
import os
import pathlib
import re
import shutil
import sys
from io import StringIO
from typing import Union, no_type_check

import treelib  # type: ignore
import yaml

import liitos.gather as gat
import liitos.placeholder as plh
import liitos.tools as too
from liitos import ENCODING, LOG_SEPARATOR, log

ALT_INJECTOR_HACK = 'INJECTED-ALT-TEXT-TO-TRIGGER-FIGURE-ENVIRONMENT-AROUND-IMAGE-IN-PANDOC'
CAP_INJECTOR_HACK = 'INJECTED-CAP-TEXT-TO-MARK-MISSING-CAPTION-IN-OUTPUT'
DOC_BASE = pathlib.Path('..', '..')
STRUCTURE_PATH = DOC_BASE / 'structure.yml'
SLASH = '/'
IMAGES_FOLDER = 'images/'
DIAGRAMS_FOLDER = 'diagrams/'

"""
```{.python .cb.run}
with open('sub/as.md') as fp:
    print(fp.read())
```
"""
READ_SLOT_FENCE_BEGIN = '```{.python .cb.run}'
READ_SLOT_CONTEXT_BEGIN = 'with open('
READ_SLOT_FENCE_END = '```'

r"""
\include{markdown_file_path}
"""
INCLUDE_SLOT = '\\include{'

"""
![Alt Text Red](images/red.png "Caption Text Red")
![Alt Text Dot Dot Lime](../images/lime.png "Caption Text Dot Dot Lime")
![Alt Text Blue](images/blue.png "Caption Text Blue")
![Alt Text Sting Red](other/images/red.png "Caption Text Sting Red")
"""
IMG_LINE_STARTSWITH = '!['
MD_IMG_PATTERN = re.compile(r'^!\[(?P<cap>[^(]*)\]\((?P<src>[^ ]+)\ *\"?(?P<alt>[^\"]*)\"?\)(?P<rest>.*)?$')
MD_IMG_PATTERN_RIGHT_SPLIT = re.compile(r'^(?P<src>[^ ]+)\ *\"?(?P<alt>[^\"]*)\"?\)(?P<rest>.*)?$')

NL = '\n'


@no_type_check
class RedirectedStdout:
    @no_type_check
    def __init__(self):
        self._stdout = None
        self._string_io = None

    @no_type_check
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._string_io = StringIO()
        return self

    @no_type_check
    def __exit__(self, type, value, traceback):
        sys.stdout = self._stdout

    @no_type_check
    def __str__(self):
        return self._string_io.getvalue()


@no_type_check
def process_approvals(aspects: dict[str, str]) -> Union[gat.Approvals, int]:
    """TODO."""
    approvals_path = DOC_BASE / aspects[gat.KEY_APPROVALS]
    if not approvals_path.is_file() or not approvals_path.stat().st_size:
        log.error(f'destructure failed to find non-empty approvals file at {approvals_path}')
        return 1
    if approvals_path.suffix.lower() not in ('.json', '.yaml', '.yml'):
        log.error(f'approvals file format per suffix ({approvals_path.suffix}) not supported')
        return 1
    approvals_channel = 'yaml' if approvals_path.suffix.lower() in ('.yaml', '.yml') else 'json'
    with open(approvals_path, 'rt', encoding=ENCODING) as handle:
        approvals = yaml.safe_load(handle) if approvals_channel == 'yaml' else json.load(handle)
    if not approvals:
        log.error(f'empty approvals file? Please add approvals to ({approvals_path})')
        return 1
    if approvals_channel == 'yaml':
        with open('approvals.yml', 'wt', encoding=ENCODING) as handle:
            yaml.dump(approvals, handle, default_flow_style=False)
    else:
        with open('approvals.json', 'wt', encoding=ENCODING) as handle:
            json.dump(approvals, handle, indent=2)
    return approvals


@no_type_check
def process_binder(aspects: dict[str, str]) -> Union[gat.Binder, int]:
    """TODO."""
    bind_path = DOC_BASE / aspects[gat.KEY_BIND]
    if not bind_path.is_file() or not bind_path.stat().st_size:
        log.error(f'destructure failed to find non-empty bind file at {bind_path}')
        return 1
    if bind_path.suffix.lower() not in ('.txt',):
        log.error(f'bind file format per suffix ({bind_path.suffix}) not supported')
        return 1
    with open(bind_path, 'rt', encoding=ENCODING) as handle:
        binder = [line.strip() for line in handle.readlines() if line.strip()]
    if not binder:
        log.error(f'empty bind file? Please add component paths to ({bind_path})')
        return 1
    with open('bind.txt', 'wt', encoding=ENCODING) as handle:
        handle.write('\n'.join(binder) + '\n')
    return binder


@no_type_check
def process_changes(aspects: dict[str, str]) -> Union[gat.Changes, int]:
    """TODO."""
    changes_path = DOC_BASE / aspects[gat.KEY_CHANGES]
    if not changes_path.is_file() or not changes_path.stat().st_size:
        log.error(f'destructure failed to find non-empty changes file at {changes_path}')
        return 1
    if changes_path.suffix.lower() not in ('.json', '.yaml', '.yml'):
        log.error(f'changes file format per suffix ({changes_path.suffix}) not supported')
        return 1
    changes_channel = 'yaml' if changes_path.suffix.lower() in ('.yaml', '.yml') else 'json'
    with open(changes_path, 'rt', encoding=ENCODING) as handle:
        changes = yaml.safe_load(handle) if changes_channel == 'yaml' else json.load(handle)
    if not changes:
        log.error(f'empty changes file? Please add changes data to ({changes_path})')
        return 1
    if changes_channel == 'yaml':
        with open('changes.yml', 'wt', encoding=ENCODING) as handle:
            yaml.dump(changes, handle, default_flow_style=False)
    else:
        with open('changes.json', 'wt', encoding=ENCODING) as handle:
            json.dump(changes, handle, indent=2)
    return changes


@no_type_check
def process_meta(aspects: dict[str, str]) -> Union[gat.Meta, int]:
    """TODO."""
    meta_path = DOC_BASE / aspects[gat.KEY_META]
    if not meta_path.is_file() or not meta_path.stat().st_size:
        log.error(f'destructure failed to find non-empty meta file at {meta_path}')
        return 1
    if meta_path.suffix.lower() not in ('.yaml', '.yml'):
        log.error(f'meta file format per suffix ({meta_path.suffix}) not supported')
        return 1
    with open(meta_path, 'rt', encoding=ENCODING) as handle:
        metadata = yaml.safe_load(handle)
    if not metadata:
        log.error(f'empty metadata file? Please add metadata to ({meta_path})')
        return 1
    if 'import' in metadata['document']:
        base_meta_path = DOC_BASE / metadata['document']['import']
        if not base_meta_path.is_file() or not base_meta_path.stat().st_size:
            log.error(
                f'metadata declares import of base data from ({base_meta_path.name})'
                f' but failed to find non-empty base file at {base_meta_path}'
            )
            return 1
        with open(base_meta_path, 'rt', encoding=ENCODING) as handle:
            base_data = yaml.safe_load(handle)
        for key, value in metadata['document']['patch'].items():
            base_data['document']['common'][key] = value
        metadata = base_data
    with open('metadata.yml', 'wt', encoding=ENCODING) as handle:
        yaml.dump(metadata, handle, default_flow_style=False)
    return metadata


@no_type_check
def parse_markdown_image(text_line: str) -> tuple[str, str, str, str]:
    """Parse a markdown image line within our conventions into caption, src, alt, and optional rest."""
    invalid_marker = ('', '', '', text_line)

    exclam = '!'
    osb = '['
    if not text_line or not text_line.startswith(f'{exclam}{osb}'):
        log.error(f'- INVALID-MD-IMG_LINE::START <<{text_line.rstrip()}>>')
        return invalid_marker

    csb = ']'
    osb_cnt = text_line.count(osb)
    csb_cnt = text_line.count(csb)
    if osb_cnt + csb_cnt < 2:
        log.error(f'- INVALID-MD-IMG_LINE::SB-TOK-CNT-LOW <<{text_line.rstrip()}>>')
        return invalid_marker
    if osb_cnt != csb_cnt:
        log.warning(f'- INCOMPLETE-MD-IMG_LINE::SB-TOK-CNT-UNBALANCED <<{text_line.rstrip()}>>')

    orb = '('
    cap_src_boundary = f'{csb}{orb}'
    if cap_src_boundary not in text_line:
        log.error(f'- INVALID-MD-IMG_LINE::CAP-SRC-BOUNDARY <<{text_line.rstrip()}>>')
        return invalid_marker

    crb = ')'
    orb_cnt = text_line.count(orb)
    crb_cnt = text_line.count(crb)
    if orb_cnt + crb_cnt < 2:
        log.error(f'- INVALID-MD-IMG_LINE::RB-TOK-CNT-LOW <<{text_line.rstrip()}>>')
        return invalid_marker
    if orb_cnt != crb_cnt:
        log.warning(f'- INCOMPLETE-MD-IMG_LINE::RB-TOK-CNT-UNBALANCED <<{text_line.rstrip()}>>')

    quo = '"'
    quo_cnt = text_line.count(quo)
    if quo_cnt < 2:
        log.warning(f'- INCOMPLETE-MD-IMG_LINE::QU-TOK-CNT-LOW <<{text_line.rstrip()}>>')
    if quo_cnt % 2:
        log.warning(f'- INCOMPLETE-MD-IMG_LINE::QU-TOK-CNT-UNBALANCED <<{text_line.rstrip()}>>')

    sp = ' '
    sp_cnt = text_line.count(sp)
    if not sp_cnt:
        log.warning(f'- INCOMPLETE-MD-IMG_LINE::SP-TOK-CNT-LOW <<{text_line.rstrip()}>>')

    dot = '.'
    sla = '/'
    abs_path_indicator = f'{csb}{orb}{sla}'
    may_have_abs_path = abs_path_indicator in text_line
    if may_have_abs_path:
        log.info(f'- SUSPICIOUS-MD-IMG_LINE::MAY-HAVE-ABS-PATH <<{text_line.rstrip()}>>')
    naive_upwards_path_indicator = f'{csb}{orb}{dot}{dot}{sla}'
    may_have_upwards_path = naive_upwards_path_indicator in text_line
    if may_have_upwards_path:
        log.info(f'- SUSPICIOUS-MD-IMG_LINE::MAY-HAVE-UPWARDS-PATH <<{text_line.rstrip()}>>')

    log.info('- parsing the markdown image text line ...')
    if orb_cnt + crb_cnt > 2 or orb_cnt != crb_cnt:
        # The regex is not safe for orb inside caption
        left, right = text_line.split(cap_src_boundary, 1)
        match_right = MD_IMG_PATTERN_RIGHT_SPLIT.match(right)
        if not match_right:
            log.error(f'- INVALID-MD-IMG_LINE::RE-MATCH-RIGHT-SPLIT-FAILED <<{text_line.rstrip()}>>')
            return invalid_marker

        parts = match_right.groupdict()
        cap = left[2:]
        if not cap:
            log.warning(f'- INCOMPLETE-MD-IMG_LINE::CAP-MISS-INJECTED <<{text_line.rstrip()}>>')
            cap = CAP_INJECTOR_HACK

        src = parts['src']
        alt = parts['alt']
        rest = parts['rest']
        if orb in alt or crb in alt:
            log.warning(f'- MAYBE-MD-IMG_LINE::ALT-TRUNCATED-PARTIAL-MATCH <<{text_line.rstrip()}>>')
            log.warning(f'  + parsed as ({cap=}, {src=}, {alt=}, {rest=}')

        return cap, src, alt, rest

    match = MD_IMG_PATTERN.match(text_line)
    if not match:
        log.error(f'- INVALID-MD-IMG_LINE::RE-MATCH-FAILED <<{text_line.rstrip()}>>')
        return invalid_marker

    parts = match.groupdict()
    cap = parts['cap']
    if not cap:
        log.warning(f'- INCOMPLETE-MD-IMG_LINE::CAP-MISS-INJECTED <<{text_line.rstrip()}>>')
        cap = CAP_INJECTOR_HACK

    src = parts['src']
    alt = parts['alt']
    rest = parts['rest']
    if orb in alt or crb in alt:
        log.warning(f'- MAYBE-MD-IMG_LINE::ALT-TRUNCATED-FULL-MATCH <<{text_line.rstrip()}>>')
        log.warning(f'  + parsed as ({cap=}, {src=}, {alt=}, {rest=}')

    return cap, src, alt, rest


@no_type_check
def adapt_image(text_line: str, collector: list[str], upstream: str, root: str) -> str:
    """YES."""
    cap, src, alt, rest = parse_markdown_image(text_line)
    if not src:
        log.error(f'parse of markdown image text line failed - empty src, and rest is <<{rest.rstrip()}>>')
        return text_line

    img_path = str((pathlib.Path(upstream).parent / src).resolve()).replace(root, '')
    collector.append(img_path)
    img_hack = img_path
    if f'/{IMAGES_FOLDER}' in img_path:
        img_hack = IMAGES_FOLDER + img_path.split(f'/{IMAGES_FOLDER}', 1)[1]
    elif f'/{DIAGRAMS_FOLDER}' in img_path:
        img_hack = DIAGRAMS_FOLDER + img_path.split(f'/{DIAGRAMS_FOLDER}', 1)[1]

    if img_hack != img_path:
        log.info(f'{img_hack} <--- OK? --- {img_path}')

    alt_text = f'"{alt}"' if alt else f'"{ALT_INJECTOR_HACK}"'
    belte_og_seler = f'![{cap}]({img_hack} {alt_text}){rest}'
    log.info(f'==> belte-og-seler: ->>{belte_og_seler}<<-')
    return belte_og_seler


@no_type_check
def harvest_include(
    text_line: str, slot: int, regions: dict[str, list[tuple[tuple[int, int], str]]], tree: treelib.Tree, parent: str
) -> None:
    """TODO."""
    include_local = text_line.split(INCLUDE_SLOT, 1)[1].rstrip('}').strip()
    include = str(pathlib.Path(parent).parent / include_local)
    regions[parent].append(((slot, slot), include))
    tree.create_node(include, include, parent=parent)


@no_type_check
def rollup(
    jobs: list[list[str]],
    docs: dict[str, list[str]],
    regions: dict[str, list[tuple[tuple[int, int], str]]],
    flat: dict[str, str],
) -> list[list[str]]:
    """TODO."""
    tackle = [those[0] for those in jobs if those and those[0] != SLASH]
    if tackle:
        log.info(f'  Insertion ongoing with parts ({", ".join(tuple(sorted(tackle)))}) remaining')
    else:
        return [[]]
    for that in tackle:
        buf = []
        for slot, line in enumerate(docs[that]):
            special = False
            the_first = False
            the_include = ''
            for pair, include in regions[that]:
                low, high = pair
                if low <= slot <= high:
                    special = True
                if low == slot:
                    the_first = True
                    the_include = include
            if not special:
                buf.append(line)
                continue
            if the_first:
                buf.append(flat[the_include])
        flat[that] = '\n'.join(buf) + '\n'

    return [[job for job in chain if job not in flat] for chain in jobs]


@no_type_check
def collect_assets(collector: list[str]) -> None:
    """TODO"""
    images = pathlib.Path(IMAGES_FOLDER)
    images.mkdir(parents=True, exist_ok=True)
    diagrams = pathlib.Path(DIAGRAMS_FOLDER)
    diagrams.mkdir(parents=True, exist_ok=True)
    for img_path in collector:
        if IMAGES_FOLDER in img_path:
            source_asset = DOC_BASE / img_path
            target_asset = images / pathlib.Path(img_path).name
            try:
                shutil.copy(source_asset, target_asset)
            except FileNotFoundError as err:
                log.error(err)
                code, msg = plh.dump_placeholder(target_asset)
                log.warning(msg) if code else log.info(msg)
            continue
        if DIAGRAMS_FOLDER in img_path:
            source_asset = DOC_BASE / img_path
            target_asset = diagrams / pathlib.Path(img_path).name
            try:
                shutil.copy(source_asset, target_asset)
            except FileNotFoundError as err:
                log.error(err)
                code, msg = plh.dump_placeholder(target_asset)
                log.warning(msg) if code else log.info(msg)


@no_type_check
def concatenate(
    doc_root: Union[str, pathlib.Path],
    structure_name: str,
    target_key: str,
    facet_key: str,
    options: dict[str, Union[bool, str]],
) -> int:
    """Later alligator."""
    log.info(LOG_SEPARATOR)
    log.info('entered concat function ...')
    target_code = target_key
    facet_code = facet_key
    if not facet_code.strip() or not target_code.strip():
        log.error(f'concatenate requires non-empty target ({target_code}) and facet ({facet_code}) codes')
        return 2

    log.info(f'parsed target ({target_code}) and facet ({facet_code}) from request')

    structure, asset_map = gat.prelude(
        doc_root=doc_root, structure_name=structure_name, target_key=target_key, facet_key=facet_key, command='concat'
    )
    log.info(f'prelude teleported processor into the document root at ({os.getcwd()}/)')
    rel_concat_folder_path = pathlib.Path('render/pdf/')
    rel_concat_folder_path.mkdir(parents=True, exist_ok=True)
    os.chdir(rel_concat_folder_path)
    log.info(f'concatenate (this processor) teleported into the render/pdf location ({os.getcwd()}/)')

    ok, aspect_map = too.load_target(target_code, facet_code)
    if not ok or not aspect_map:
        return 0 if ok else 1

    approvals = process_approvals(aspect_map)
    if isinstance(approvals, int):
        return 1
    binder = process_binder(aspect_map)
    if isinstance(binder, int):
        return 1
    changes = process_changes(aspect_map)
    if isinstance(changes, int):
        return 1
    metadata = process_meta(aspect_map)
    if isinstance(metadata, int):
        return 1

    root = SLASH
    root_path = str(pathlib.Path.cwd().resolve()).rstrip(SLASH) + SLASH
    tree = treelib.Tree()
    tree.create_node(root, root)
    documents = {}
    insert_regions = {}
    img_collector = []
    log.info(LOG_SEPARATOR)
    log.info('processing binder ...')
    for entry in binder:
        ref_path = DOC_BASE / entry
        log.debug(f'- {entry} as {ref_path}')
        with open(ref_path, 'rt', encoding=ENCODING) as handle:
            documents[entry] = [line.rstrip() for line in handle.readlines()]
        insert_regions[entry] = []
        in_region = False
        begin, end = 0, 0
        include = ''
        tree.create_node(entry, entry, parent=root)
        for slot, line in enumerate(documents[entry]):
            if line.startswith(IMG_LINE_STARTSWITH):
                documents[entry][slot] = adapt_image(line, img_collector, entry, root_path)
            log.debug(f'{slot :02d}|{line.rstrip()}')
            if not in_region:
                if line.startswith(READ_SLOT_FENCE_BEGIN):
                    in_region = True
                    begin = slot
                    continue
                if line.startswith(INCLUDE_SLOT):
                    include = line.split(INCLUDE_SLOT, 1)[1].rstrip('}').strip()
                    insert_regions[entry].append(((slot, slot), include))
                    tree.create_node(include, include, parent=entry)
                    include = ''
                    continue
            if in_region:
                if line.startswith(READ_SLOT_CONTEXT_BEGIN):
                    include = line.replace(READ_SLOT_CONTEXT_BEGIN, '').split(')', 1)[0].strip("'").strip('"')
                elif line.startswith(READ_SLOT_FENCE_END):
                    end = slot
                    insert_regions[entry].append(((begin, end), include))
                    tree.create_node(include, include, parent=entry)
                    in_region = False
                    begin, end = 0, 0
                    include = ''

        for coords, include in insert_regions[entry]:  # include is anchored on DOC_BASE
            ref_path = DOC_BASE / include
            with open(ref_path, 'rt', encoding=ENCODING) as handle:
                documents[include] = [line.rstrip() for line in handle.readlines()]
            insert_regions[include] = []
            in_region = False
            begin, end = 0, 0
            sub_include = ''
            for slot, line in enumerate(documents[include]):
                if line.startswith(IMG_LINE_STARTSWITH):
                    documents[include][slot] = adapt_image(line, img_collector, include, root_path)
                log.debug(f'{slot :02d}|{line.rstrip()}')
                if not in_region:
                    if line.startswith(READ_SLOT_FENCE_BEGIN):
                        in_region = True
                        begin = slot
                        continue
                    if line.startswith(INCLUDE_SLOT):
                        harvest_include(line, slot, insert_regions, tree, include)
                        continue
                if in_region:
                    if line.startswith(READ_SLOT_CONTEXT_BEGIN):
                        sub_include = line.replace(READ_SLOT_CONTEXT_BEGIN, '').split(')', 1)[0].strip("'").strip('"')
                        sub_include = str(pathlib.Path(include).parent / sub_include)
                    elif line.startswith(READ_SLOT_FENCE_END):
                        end = slot
                        insert_regions[include].append(((begin, end), sub_include))
                        tree.create_node(sub_include, sub_include, parent=include)
                        in_region = False
                        begin, end = 0, 0
                        sub_include = ''

            for coords, sub_include in insert_regions[include]:
                ref_path = DOC_BASE / sub_include
                with open(ref_path, 'rt', encoding=ENCODING) as handle:
                    documents[sub_include] = [line.rstrip() for line in handle.readlines()]
                insert_regions[sub_include] = []
                in_region = False
                begin, end = 0, 0
                sub_sub_include = ''
                for slot, line in enumerate(documents[sub_include]):
                    if line.startswith(IMG_LINE_STARTSWITH):
                        documents[sub_include][slot] = adapt_image(line, img_collector, sub_include, root_path)
                    log.debug(f'{slot :02d}|{line.rstrip()}')
                    if not in_region:
                        if line.startswith(READ_SLOT_FENCE_BEGIN):
                            in_region = True
                            begin = slot
                            continue
                        if line.startswith(INCLUDE_SLOT):
                            harvest_include(line, slot, insert_regions, tree, sub_include)
                            continue
                    if in_region:
                        if line.startswith(READ_SLOT_CONTEXT_BEGIN):
                            sub_sub_include = (
                                line.replace(READ_SLOT_CONTEXT_BEGIN, '').split(')', 1)[0].strip("'").strip('"')
                            )
                            sub_sub_include = str(pathlib.Path(sub_include).parent / sub_sub_include)
                        elif line.startswith(READ_SLOT_FENCE_END):
                            end = slot
                            insert_regions[sub_include].append(((begin, end), sub_sub_include))
                            tree.create_node(sub_sub_include, sub_sub_include, parent=sub_include)
                            in_region = False
                            begin, end = 0, 0
                            sub_sub_include = ''

                for coords, sub_sub_include in insert_regions[sub_include]:
                    ref_path = DOC_BASE / sub_sub_include
                    with open(ref_path, 'rt', encoding=ENCODING) as handle:
                        documents[sub_sub_include] = [line.rstrip() for line in handle.readlines()]
                    insert_regions[sub_sub_include] = []
                    in_region = False
                    begin, end = 0, 0
                    sub_sub_sub_include = ''
                    for slot, line in enumerate(documents[sub_sub_include]):
                        if line.startswith(IMG_LINE_STARTSWITH):
                            documents[sub_sub_include][slot] = adapt_image(
                                line, img_collector, sub_sub_include, root_path
                            )
                        log.debug(f'{slot :02d}|{line.rstrip()}')
                        if not in_region:
                            if line.startswith(READ_SLOT_FENCE_BEGIN):
                                in_region = True
                                begin = slot
                                continue
                            if line.startswith(INCLUDE_SLOT):
                                harvest_include(line, slot, insert_regions, tree, sub_sub_include)
                                continue
                        if in_region:
                            if line.startswith(READ_SLOT_CONTEXT_BEGIN):
                                sub_sub_sub_include = (
                                    line.replace(READ_SLOT_CONTEXT_BEGIN, '').split(')', 1)[0].strip("'").strip('"')
                                )
                                sub_sub_sub_include = str(pathlib.Path(sub_sub_include).parent / sub_sub_sub_include)
                            elif line.startswith(READ_SLOT_FENCE_END):
                                end = slot
                                insert_regions[sub_sub_include].append(((begin, end), sub_sub_sub_include))
                                tree.create_node(sub_sub_sub_include, sub_sub_sub_include, parent=sub_sub_include)
                                in_region = False
                                begin, end = 0, 0
                                sub_sub_sub_include = ''

                    for coords, sub_sub_sub_include in insert_regions[sub_include]:
                        ref_path = DOC_BASE / sub_sub_sub_include
                        with open(ref_path, 'rt', encoding=ENCODING) as handle:
                            documents[sub_sub_sub_include] = [line.rstrip() for line in handle.readlines()]
                        insert_regions[sub_sub_sub_include] = []
                        in_region = False
                        begin, end = 0, 0
                        sub_sub_sub_sub_include = ''
                        for slot, line in enumerate(documents[sub_sub_sub_include]):
                            if line.startswith(IMG_LINE_STARTSWITH):
                                documents[sub_sub_sub_include][slot] = adapt_image(
                                    line, img_collector, sub_sub_sub_include, root_path
                                )
                            log.debug(f'{slot :02d}|{line.rstrip()}')
                            if not in_region:
                                if line.startswith(READ_SLOT_FENCE_BEGIN):
                                    in_region = True
                                    begin = slot
                                    continue
                                if line.startswith(INCLUDE_SLOT):
                                    harvest_include(line, slot, insert_regions, tree, sub_sub_sub_include)
                                    continue
                            if in_region:
                                if line.startswith(READ_SLOT_CONTEXT_BEGIN):
                                    sub_sub_sub_sub_include = (
                                        line.replace(READ_SLOT_CONTEXT_BEGIN, '').split(')', 1)[0].strip("'").strip('"')
                                    )
                                    sub_sub_sub_sub_include = str(
                                        pathlib.Path(sub_sub_sub_include).parent / sub_sub_sub_sub_include
                                    )
                                elif line.startswith(READ_SLOT_FENCE_END):
                                    end = slot
                                    insert_regions[sub_sub_sub_include].append(((begin, end), sub_sub_sub_include))
                                    tree.create_node(
                                        sub_sub_sub_sub_include, sub_sub_sub_sub_include, parent=sub_sub_sub_include
                                    )
                                    in_region = False
                                    begin, end = 0, 0
                                    sub_sub_sub_sub_include = ''

    top_down_paths = tree.paths_to_leaves()
    bottom_up_paths = [list(reversed(td_p)) for td_p in top_down_paths]
    log.info(LOG_SEPARATOR)
    log.info('resulting tree:')
    for edge in str(tree).split(NL):
        log.info(edge)

    log.info(LOG_SEPARATOR)
    log.info(f'provisioning chains for the {len(bottom_up_paths)} bottom up leaf paths:')
    for num, leaf_path in enumerate(bottom_up_paths):
        the_way_up = f'|-> {leaf_path[0]}' if len(leaf_path) == 1 else f'{" -> ".join(leaf_path)}'
        log.info(f'{num :2d}: {the_way_up}')

    concat = {}
    log.info(LOG_SEPARATOR)
    log.info(f'dependencies for the {len(insert_regions)} document parts:')
    for key, regions in insert_regions.items():
        num_in = len(regions)
        dashes = '-' * num_in
        incl_disp = f'( {num_in} include{"" if num_in == 1 else "s"} )'
        indicator = '(no includes)' if not regions else f'<{dashes + incl_disp + dashes}'
        log.info(f'- part {key} {indicator}')
        for region in regions:
            between = f'between lines {region[0][0] :3d} and {region[0][1] :3d}'
            insert = f'include fragment {region[1]}'
            log.info(f'  + {between} {insert}')
        if not regions:  # No includes
            concat[key] = '\n'.join(documents[key]) + '\n'
            log.info(f'  * did concat {key} document for insertion')

    chains = [leaf_path for leaf_path in bottom_up_paths]
    log.info(LOG_SEPARATOR)
    log.info(f'starting insertions bottom up for the {len(chains)} inclusion chains:')
    todo = [[job for job in chain if job not in concat] for chain in chains]
    while todo != [[]]:
        todo = rollup(todo, documents, insert_regions, concat)

    log.info(LOG_SEPARATOR)
    log.info('writing final concat markdown to document.md')
    with open('document.md', 'wt', encoding=ENCODING) as handle:
        handle.write('\n'.join(concat[bind] for bind in binder) + '\n')

    log.info(LOG_SEPARATOR)
    log.info('collecting assets (images and diagrams)')
    collect_assets(img_collector)
    log.info(LOG_SEPARATOR)
    log.info(f'concat result document (document.md) and artifacts are within folder ({os.getcwd()}/)')
    log.info(LOG_SEPARATOR)
    log.info('processing complete - SUCCESS')
    log.info(LOG_SEPARATOR)
    return 0



"""Apply all pairs in patches to incoming."""

from collections.abc import Iterable

from liitos import log


def apply(patches: list[tuple[str, str]], incoming: Iterable[str]) -> list[str]:
    """Later alligator."""
    outgoing = [line for line in incoming]

    log.info(f'applying patches to {len(outgoing)} lines of text')
    for this, that in patches:
        log.info(f' - trying any ({this}) --> ({that}) ...')
        for n, text in enumerate(outgoing):
            if this in text:
                log.info(f'   + found match ({text})')
                outgoing[n] = text.replace(this, that)

    return outgoing



import os
import pathlib

import liitos.meta as meta

EXAMPLE_DEEP_DOC_ROOT = pathlib.Path('example', 'deep')


def test_weave_meta_part_proprietary_information_on_empty_ok():
    assert meta.weave_meta_part_proprietary_information({}, '') == ''


def test_weave_meta_part_proprietary_information_on_value_slot_ok():
    assert meta.weave_meta_part_proprietary_information({}, '-VALUE.SLOT+') == '-This is a notice.\n+'  # noqa


def test_meta_dispatch_no_match_let_pass_empty():
    assert meta.weave_meta_meta({'document': {'common': {}}}, []) == []


def test_meta_dispatch_no_match_let_pass():
    assert meta.weave_meta_meta({'document': {'common': {}}}, ['no-match']) == ['no-match', '\n']


def test_meta_dispatch():
    dispatch = {
        '%%_PATCH_%_HEADER_%_TITLE_%%': meta.weave_meta_part_header_title,
        '%%_PATCH_%_MAIN_%_TITLE_%%': meta.weave_meta_part_title,
        '%%_PATCH_%_SUB_%_TITLE_%%': meta.weave_meta_part_sub_title,
        '%%_PATCH_%_TYPE_%%': meta.weave_meta_part_header_type,
        '%%_PATCH_%_ID_%%': meta.weave_meta_part_header_id,
        '%%_PATCH_%_ISSUE_%%': meta.weave_meta_part_issue,
        '%%_PATCH_%_REVISION_%%': meta.weave_meta_part_revision,
        '%%_PATCH_%_DATE_%%': meta.weave_meta_part_header_date,
        '%%_PATCH_%_FRAME_%_NOTE_%%': meta.weave_meta_part_footer_frame_note,
        '%%_PATCH_%_FOOT_%_PAGE_%_COUNTER_%_LABEL_%%': meta.weave_meta_part_footer_page_number_prefix,
        '%%_PATCH_%_CHANGELOG_%_ISSUE_%_LABEL_%%': meta.weave_meta_part_change_log_issue_label,
        '%%_PATCH_%_CHANGELOG_%_REVISION_%_LABEL_%%': meta.weave_meta_part_change_log_revision_label,
        '%%_PATCH_%_CHANGELOG_%_DATE_%_LABEL_%%': meta.weave_meta_part_change_log_date_label,
        '%%_PATCH_%_CHANGELOG_%_AUTHOR_%_LABEL_%%': meta.weave_meta_part_change_log_author_label,
        '%%_PATCH_%_CHANGELOG_%_DESCRIPTION_%_LABEL_%%': meta.weave_meta_part_change_log_description_label,
        '%%_PATCH_%_APPROVALS_%_ADJUSTABLE_%_VERTICAL_%_SPACE_%%': meta.weave_meta_part_approvals_adjustable_vertical_space,  # noqa
        '%%_PATCH_%_APPROVALS_%_ROLE_%_LABEL_%%': meta.weave_meta_part_approvals_role_label,
        '%%_PATCH_%_APPROVALS_%_NAME_%_LABEL_%%': meta.weave_meta_part_approvals_name_label,
        '%%_PATCH_%_APPROVALS_%_DATE_%_AND_%_SIGNATURE_%_LABEL_%%': meta.weave_meta_part_approvals_date_and_signature_label,  # noqa
        '%%_PATCH_%_BLURB_%_ADJUSTABLE_%_VERTICAL_%_SPACE_%%': meta.weave_meta_part_proprietary_information_adjustable_vertical_space,  # noqa
        '%%_PATCH_%_ISSUE_%_REVISION_%_COMBINED_%%': meta.weave_meta_part_header_issue_revision_combined,
        '%%_PATCH_%_PROPRIETARY_%_INFORMATION_%_LABEL_%%': meta.weave_meta_part_proprietary_information,
    }
    expected = {
        '%%_PATCH_%_HEADER_%_TITLE_%%': '',
        '%%_PATCH_%_MAIN_%_TITLE_%%': '',
        '%%_PATCH_%_SUB_%_TITLE_%%': ' ',
        '%%_PATCH_%_TYPE_%%': 'Engineering Document',
        '%%_PATCH_%_ID_%%': 'N/A',
        '%%_PATCH_%_ISSUE_%%': '01',
        '%%_PATCH_%_REVISION_%%': '00',
        '%%_PATCH_%_DATE_%%': '01 FEB 2345',
        '%%_PATCH_%_FRAME_%_NOTE_%%': ' ',
        '%%_PATCH_%_FOOT_%_PAGE_%_COUNTER_%_LABEL_%%': 'Page',
        '%%_PATCH_%_CHANGELOG_%_ISSUE_%_LABEL_%%': 'Iss.',
        '%%_PATCH_%_CHANGELOG_%_REVISION_%_LABEL_%%': 'Rev.',
        '%%_PATCH_%_CHANGELOG_%_DATE_%_LABEL_%%': 'Date',
        '%%_PATCH_%_CHANGELOG_%_AUTHOR_%_LABEL_%%': 'Author',
        '%%_PATCH_%_CHANGELOG_%_DESCRIPTION_%_LABEL_%%': 'Description',
        '%%_PATCH_%_APPROVALS_%_ADJUSTABLE_%_VERTICAL_%_SPACE_%%': '2.5em',
        '%%_PATCH_%_APPROVALS_%_ROLE_%_LABEL_%%': 'Approvals',
        '%%_PATCH_%_APPROVALS_%_NAME_%_LABEL_%%': 'Name',
        '%%_PATCH_%_APPROVALS_%_DATE_%_AND_%_SIGNATURE_%_LABEL_%%': 'Date and Signature',
        '%%_PATCH_%_BLURB_%_ADJUSTABLE_%_VERTICAL_%_SPACE_%%': '-0em',
        '%%_PATCH_%_ISSUE_%_REVISION_%_COMBINED_%%': r'Iss \theMetaIssCode, Rev \theMetaRevCode',
        '%%_PATCH_%_PROPRIETARY_%_INFORMATION_%_LABEL_%%': 'This is a notice.',
    }
    mapper = {'title': '', 'header_date': '01 FEB 2345'}
    value_slot_container = '-VALUE.SLOT+'
    wrapper = {'document': {'common': {**mapper}}}
    for trigger, weaver in dispatch.items():
        hack = '\n' if trigger == '%%_PATCH_%_PROPRIETARY_%_INFORMATION_%_LABEL_%%' else ''
        assert weaver(mapper, value_slot_container) == f'-{expected[trigger]}{hack}+'
        value_wrapper = f'-VALUE.SLOT+{trigger}'
        assert meta.weave_meta_meta(wrapper, [value_wrapper]) == [f'-{expected[trigger]}{hack}+{trigger}', '\n']


def test_meta_dispatch_explicit():
    dispatch = {
        '%%_PATCH_%_HEADER_%_TITLE_%%': meta.weave_meta_part_header_title,
        '%%_PATCH_%_MAIN_%_TITLE_%%': meta.weave_meta_part_title,
        '%%_PATCH_%_SUB_%_TITLE_%%': meta.weave_meta_part_sub_title,
        '%%_PATCH_%_TYPE_%%': meta.weave_meta_part_header_type,
        '%%_PATCH_%_ID_%%': meta.weave_meta_part_header_id,
        '%%_PATCH_%_ISSUE_%%': meta.weave_meta_part_issue,
        '%%_PATCH_%_REVISION_%%': meta.weave_meta_part_revision,
        '%%_PATCH_%_DATE_%%': meta.weave_meta_part_header_date,
        '%%_PATCH_%_FRAME_%_NOTE_%%': meta.weave_meta_part_footer_frame_note,
        '%%_PATCH_%_FOOT_%_PAGE_%_COUNTER_%_LABEL_%%': meta.weave_meta_part_footer_page_number_prefix,
        '%%_PATCH_%_CHANGELOG_%_ISSUE_%_LABEL_%%': meta.weave_meta_part_change_log_issue_label,
        '%%_PATCH_%_CHANGELOG_%_REVISION_%_LABEL_%%': meta.weave_meta_part_change_log_revision_label,
        '%%_PATCH_%_CHANGELOG_%_DATE_%_LABEL_%%': meta.weave_meta_part_change_log_date_label,
        '%%_PATCH_%_CHANGELOG_%_AUTHOR_%_LABEL_%%': meta.weave_meta_part_change_log_author_label,
        '%%_PATCH_%_CHANGELOG_%_DESCRIPTION_%_LABEL_%%': meta.weave_meta_part_change_log_description_label,
        '%%_PATCH_%_APPROVALS_%_ADJUSTABLE_%_VERTICAL_%_SPACE_%%': meta.weave_meta_part_approvals_adjustable_vertical_space,  # noqa
        '%%_PATCH_%_APPROVALS_%_ROLE_%_LABEL_%%': meta.weave_meta_part_approvals_role_label,
        '%%_PATCH_%_APPROVALS_%_NAME_%_LABEL_%%': meta.weave_meta_part_approvals_name_label,
        '%%_PATCH_%_APPROVALS_%_DATE_%_AND_%_SIGNATURE_%_LABEL_%%': meta.weave_meta_part_approvals_date_and_signature_label,  # noqa
        '%%_PATCH_%_ISSUE_%_REVISION_%_COMBINED_%%': meta.weave_meta_part_header_issue_revision_combined,
        '%%_PATCH_%_PROPRIETARY_%_INFORMATION_%_LABEL_%%': meta.weave_meta_part_proprietary_information,
    }
    mapper = {
        'header_title': 'Ttt Tt',
        'title': 'Ttt Tt Tt',
        'sub_title': 'The Deep Spec',
        'header_type': 'Engineering Document',
        'header_id': 'MMI',
        'issue': '01',
        'revision': '00',
        'header_date': '01 FEB 2345',
        'footer_frame_note': 'VERY CONSEQUENTIAL',
        'footer_page_number_prefix': 'Page',
        'change_log_issue_label': 'Iss.',
        'change_log_revision_label': 'Rev.',
        'change_log_date_label': 'Date',
        'change_log_author_label': 'Author',
        'change_log_description_label': 'Description',
        'approvals_adjustable_vertical_space': '2.5em',
        'approvals_role_label': 'Approvals',
        'approvals_name_label': 'Name',
        'approvals_date_and_signature_label': 'Date and Signature',
        'header_issue_revision_combined': 'combined',
        'proprietary_information': 'test',
    }
    expected = {
        '%%_PATCH_%_HEADER_%_TITLE_%%': mapper['header_title'],
        '%%_PATCH_%_MAIN_%_TITLE_%%': mapper['title'],
        '%%_PATCH_%_SUB_%_TITLE_%%': mapper['sub_title'],
        '%%_PATCH_%_TYPE_%%': mapper['header_type'],
        '%%_PATCH_%_ID_%%': mapper['header_id'],
        '%%_PATCH_%_ISSUE_%%': mapper['issue'],
        '%%_PATCH_%_REVISION_%%': mapper['revision'],
        '%%_PATCH_%_DATE_%%': mapper['header_date'],
        '%%_PATCH_%_FRAME_%_NOTE_%%': mapper['footer_frame_note'],
        '%%_PATCH_%_FOOT_%_PAGE_%_COUNTER_%_LABEL_%%': mapper['footer_page_number_prefix'],
        '%%_PATCH_%_CHANGELOG_%_ISSUE_%_LABEL_%%': mapper['change_log_issue_label'],
        '%%_PATCH_%_CHANGELOG_%_REVISION_%_LABEL_%%': mapper['change_log_revision_label'],
        '%%_PATCH_%_CHANGELOG_%_DATE_%_LABEL_%%': mapper['change_log_date_label'],
        '%%_PATCH_%_CHANGELOG_%_AUTHOR_%_LABEL_%%': mapper['change_log_author_label'],
        '%%_PATCH_%_CHANGELOG_%_DESCRIPTION_%_LABEL_%%': mapper['change_log_description_label'],
        '%%_PATCH_%_APPROVALS_%_ADJUSTABLE_%_VERTICAL_%_SPACE_%%': mapper['approvals_adjustable_vertical_space'],
        '%%_PATCH_%_APPROVALS_%_ROLE_%_LABEL_%%': mapper['approvals_role_label'],
        '%%_PATCH_%_APPROVALS_%_NAME_%_LABEL_%%': mapper['approvals_name_label'],
        '%%_PATCH_%_APPROVALS_%_DATE_%_AND_%_SIGNATURE_%_LABEL_%%': mapper['approvals_date_and_signature_label'],
        '%%_PATCH_%_ISSUE_%_REVISION_%_COMBINED_%%': mapper['header_issue_revision_combined'],
        '%%_PATCH_%_PROPRIETARY_%_INFORMATION_%_LABEL_%%': mapper['proprietary_information'],
    }

    value_slot_container = '-VALUE.SLOT+'
    wrapper = {'document': {'common': {**mapper}}}
    for trigger, weaver in dispatch.items():
        assert weaver(mapper, value_slot_container) == f'-{expected[trigger]}+'
        value_wrapper = f'-VALUE.SLOT+{trigger}'
        assert meta.weave_meta_meta(wrapper, [value_wrapper]) == [f'-{expected[trigger]}+{trigger}', '\n']


def test_driver_dispatch_no_match_let_pass():
    assert meta.weave_meta_driver({'document': {'common': {}}}, ['no-match']) == ['no-match', '\n']


def test_driver_dispatch_no_match_let_pass_empty():
    assert meta.weave_meta_driver({'document': {'common': {}}}, []) == []


def test_driver_dispatch():
    dispatch = {
        '%%_PATCH_%_TOC_%_LEVEL_%%': meta.weave_driver_toc_level,
        '%%_PATCH_%_LOF_%%': meta.weave_driver_list_of_figures,
        '%%_PATCH_%_LOT_%%': meta.weave_driver_list_of_tables,
    }
    expected = {
        '%%_PATCH_%_TOC_%_LEVEL_%%': '2',
        '%%_PATCH_%_LOF_%%': '%',
        '%%_PATCH_%_LOT_%%': '%',
    }
    mapper = {'title': '', 'header_date': '01 FEB 2345'}
    value_slot_container = '-VALUE.SLOT+'
    wrapper = {'document': {'common': {**mapper}}}
    for trigger, weaver in dispatch.items():
        assert weaver(mapper, value_slot_container) == f'-{expected[trigger]}+'
        value_wrapper = f'-VALUE.SLOT+{trigger}'
        assert meta.weave_meta_driver(wrapper, [value_wrapper]) == [f'-{expected[trigger]}+{trigger}', '\n']


def test_driver_dispatch_explicit():
    dispatch = {
        '%%_PATCH_%_TOC_%_LEVEL_%%': meta.weave_driver_toc_level,
        '%%_PATCH_%_LOF_%%': meta.weave_driver_list_of_figures,
        '%%_PATCH_%_LOT_%%': meta.weave_driver_list_of_tables,
    }
    expected = {
        # WARNING ignored toc level (5) set to default (2) - expected value 0 < toc_level < 5:
        '%%_PATCH_%_TOC_%_LEVEL_%%': '2',
        '%%_PATCH_%_LOF_%%': '',
        '%%_PATCH_%_LOT_%%': '%',
    }
    mapper = {'toc_level': 5, 'list_of_figures': '', 'list_of_tables': '%'}
    value_slot_container = '-VALUE.SLOT+'
    wrapper = {'document': {'common': {**mapper}}}
    for trigger, weaver in dispatch.items():
        assert weaver(mapper, value_slot_container) == f'-{expected[trigger]}+'
        value_wrapper = f'-VALUE.SLOT+{trigger}'
        assert meta.weave_meta_driver(wrapper, [value_wrapper]) == [f'-{expected[trigger]}+{trigger}', '\n']
    # Parked here for now: Not a number for toc level
    assert (
        meta.weave_driver_toc_level({'toc_level': 'x'}, '-VALUE.SLOT+%%_PATCH_%_TOC_%_LEVEL_%%')
        == '-2+%%_PATCH_%_TOC_%_LEVEL_%%'
    )
    # Parked here for now: Unaccepted value for switching of list of figures and tables
    assert (
        meta.weave_driver_list_of_figures({'list_of_figures': 'x'}, '-VALUE.SLOT+%%_PATCH_%_LOF_%%')
        == '-%+%%_PATCH_%_LOF_%%'
    )
    assert (
        meta.weave_driver_list_of_tables({'list_of_tables': 'x'}, '-VALUE.SLOT+%%_PATCH_%_LOT_%%')
        == '-%+%%_PATCH_%_LOT_%%'
    )


def test_setup_dispatch_no_match_let_pass_empty():
    assert meta.weave_meta_setup({'document': {'common': {}}}, []) == []


def test_setup_dispatch_no_match_let_pass():
    assert meta.weave_meta_setup({'document': {'common': {}}}, ['no-match']) == ['no-match', '\n']


def test_setup_dispatch():
    dispatch = {
        '%%_PATCH_%_FONT_%_PATH_%%': meta.weave_setup_font_path,
        '%%_PATCH_%_FONT_%_SUFFIX_%%': meta.weave_setup_font_suffix,
        '%%_PATCH_%_BOLD_%_FONT_%%': meta.weave_setup_bold_font,
        '%%_PATCH_%_ITALIC_%_FONT_%%': meta.weave_setup_italic_font,
        '%%_PATCH_%_BOLDITALIC_%_FONT_%%': meta.weave_setup_bold_italic_font,
        '%%_PATCH_%_MAIN_%_FONT_%%': meta.weave_setup_main_font,
        '%%_PATCH_%_FIXED_%_FONT_%_PACKAGE_%%': meta.weave_setup_fixed_font_package,
        '%%_PATCH_%_CODE_%_FONTSIZE_%%': meta.weave_setup_code_fontsize,
        '%%_PATCH_%_CHOSEN_%_LOGO_%%': meta.weave_setup_chosen_logo,
    }
    d = {**meta.WEAVE_DEFAULTS}
    expected = {
        '%%_PATCH_%_FONT_%_PATH_%%': d['font_path'],
        '%%_PATCH_%_FONT_%_SUFFIX_%%': d['font_suffix'],
        '%%_PATCH_%_BOLD_%_FONT_%%': d['bold_font'],
        '%%_PATCH_%_ITALIC_%_FONT_%%': d['italic_font'],
        '%%_PATCH_%_BOLDITALIC_%_FONT_%%': d['bold_italic_font'],
        '%%_PATCH_%_MAIN_%_FONT_%%': d['main_font'],
        '%%_PATCH_%_FIXED_%_FONT_%_PACKAGE_%%': d['fixed_font_package'],
        '%%_PATCH_%_CODE_%_FONTSIZE_%%': d['code_fontsize'],
        '%%_PATCH_%_CHOSEN_%_LOGO_%%': d['chosen_logo'],
    }
    mapper = {'title': '', 'header_date': '01 FEB 2345'}
    value_slot_container = '-VALUE.SLOT+'
    wrapper = {'document': {'common': {**mapper}}}
    for trigger, weaver in dispatch.items():
        assert weaver(mapper, value_slot_container) == f'-{expected[trigger]}+'
        value_wrapper = f'-VALUE.SLOT+{trigger}'
        assert meta.weave_meta_setup(wrapper, [value_wrapper]) == [f'-{expected[trigger]}+{trigger}', '\n']


def test_setup_dispatch_explicit():
    dispatch = {
        '%%_PATCH_%_FONT_%_PATH_%%': meta.weave_setup_font_path,
        '%%_PATCH_%_FONT_%_SUFFIX_%%': meta.weave_setup_font_suffix,
        '%%_PATCH_%_BOLD_%_FONT_%%': meta.weave_setup_bold_font,
        '%%_PATCH_%_ITALIC_%_FONT_%%': meta.weave_setup_italic_font,
        '%%_PATCH_%_BOLDITALIC_%_FONT_%%': meta.weave_setup_bold_italic_font,
        '%%_PATCH_%_MAIN_%_FONT_%%': meta.weave_setup_main_font,
        '%%_PATCH_%_FIXED_%_FONT_%_PACKAGE_%%': meta.weave_setup_fixed_font_package,
        '%%_PATCH_%_CODE_%_FONTSIZE_%%': meta.weave_setup_code_fontsize,
        '%%_PATCH_%_CHOSEN_%_LOGO_%%': meta.weave_setup_chosen_logo,
    }
    d = {**meta.WEAVE_DEFAULTS}
    expected = {
        '%%_PATCH_%_FONT_%_PATH_%%': d['font_path'],
        '%%_PATCH_%_FONT_%_SUFFIX_%%': d['font_suffix'],
        '%%_PATCH_%_BOLD_%_FONT_%%': d['bold_font'],
        '%%_PATCH_%_ITALIC_%_FONT_%%': d['italic_font'],
        '%%_PATCH_%_BOLDITALIC_%_FONT_%%': d['bold_italic_font'],
        '%%_PATCH_%_MAIN_%_FONT_%%': d['main_font'],
        '%%_PATCH_%_FIXED_%_FONT_%_PACKAGE_%%': d['fixed_font_package'],
        '%%_PATCH_%_CODE_%_FONTSIZE_%%': d['code_fontsize'],
        '%%_PATCH_%_CHOSEN_%_LOGO_%%': d['chosen_logo'],
    }
    mapper = {**d}
    value_slot_container = '-VALUE.SLOT+'
    wrapper = {'document': {'common': {**mapper}}}
    for trigger, weaver in dispatch.items():
        assert weaver(mapper, value_slot_container) == f'-{expected[trigger]}+'
        value_wrapper = f'-VALUE.SLOT+{trigger}'
        assert meta.weave_meta_setup(wrapper, [value_wrapper]) == [f'-{expected[trigger]}+{trigger}', '\n']


def test_ren_der():
    parameters = {
        'doc_root': EXAMPLE_DEEP_DOC_ROOT,
        'structure_name': 'structure.yml',
        'target_key': 'prod_kind',
        'facet_key': 'deep',
        'options': {},
    }
    restore = os.getcwd()
    assert meta.weave(**parameters) == 0
    os.chdir(restore)
